"""Testing for Bag-of-SFA Symbols."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from scipy.sparse import csr_matrix
from sklearn.feature_extraction.text import CountVectorizer
from pyts.transformation import BOSS
from pyts.approximation import SymbolicFourierApproximation


n_samples, n_timestamps, n_classes = 8, 200, 2
rng = np.random.RandomState(42)
X = rng.randn(n_samples, n_timestamps)
y = rng.randint(n_classes, size=n_samples)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'word_size': "3"}, TypeError, "'word_size' must be an integer."),

     ({'window_size': {}}, TypeError,
      "'window_size' must be an integer or a float."),

     ({'window_step': {}}, TypeError,
      "'window_step' must be an integer or a float."),

     ({'word_size': 0}, ValueError, "'word_size' must be a positive integer."),

     ({'window_size': 0, 'drop_sum': True}, ValueError,
      "If 'window_size' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to (n_timestamps - 1) if 'drop_sum=True'."),

     ({'window_size': n_timestamps, 'drop_sum': True}, ValueError,
      "If 'window_size' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to (n_timestamps - 1) if 'drop_sum=True'."),

     ({'window_size': 0}, ValueError,
      "If 'window_size' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps if 'drop_sum=False'."),

     ({'window_size': n_timestamps + 1}, ValueError,
      "If 'window_size' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps if 'drop_sum=False'."),

     ({'window_size': 1.5}, ValueError,
      "If 'window_size' is a float, it must be greater than 0 and lower than "
      "or equal to 1."),

     ({'window_step': 0}, ValueError,
      "If 'window_step' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps."),

     ({'window_step': n_timestamps + 1}, ValueError,
      "If 'window_step' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps."),

     ({'window_step': 0.}, ValueError,
      "If 'window_step' is a float, it must be greater than 0 and lower than "
      "or equal to 1."),

     ({'window_step': 1.2}, ValueError,
      "If 'window_step' is a float, it must be greater than 0 and lower than "
      "or equal to 1."),

     ({'window_size': 4, 'drop_sum': True}, ValueError,
      "'word_size' must be lower than or equal to (window_size - 1) if "
      "'drop_sum=True'."),

     ({'window_size': 3}, ValueError,
      "'word_size' must be lower than or equal to window_size if "
      "'drop_sum=False'.")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    boss = BOSS(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        boss.fit(X, y)


@pytest.mark.parametrize(
    'sparse, instance', [(True, csr_matrix), (False, np.ndarray)])
def test_sparse_dense(sparse, instance):
    """Test that the expected type is returned."""
    weasel = BOSS(sparse=sparse)
    assert isinstance(weasel.fit(X, y).transform(X), instance)
    assert isinstance(weasel.fit_transform(X, y), instance)


def test_accurate_results_without_numerosity_reduction():
    """Test that the actual results are the expected ones."""
    boss = BOSS(
        word_size=4, n_bins=3, window_size=100, window_step=100,
        anova=False, drop_sum=False, norm_mean=False, norm_std=False,
        strategy='quantile', alphabet=None, numerosity_reduction=False
    )

    X_windowed = X.reshape(8, 2, 100).reshape(16, 100)
    sfa = SymbolicFourierApproximation(
        n_coefs=4, drop_sum=False, anova=False, norm_mean=False,
        norm_std=False, n_bins=3, strategy='quantile', alphabet=None
    )
    y_repeated = np.repeat(y, 2)
    X_sfa = sfa.fit_transform(X_windowed, y_repeated)
    X_word = np.asarray([''.join(X_sfa[i]) for i in range(16)])
    X_word = X_word.reshape(8, 2)
    X_bow = np.asarray([' '.join(X_word[i]) for i in range(8)])

    vectorizer = CountVectorizer()
    arr_desired = vectorizer.fit_transform(X_bow).toarray()
    vocabulary_desired = {value: key for key, value in
                          vectorizer.vocabulary_.items()}

    arr_actual = boss.fit_transform(X, y).toarray()
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0)
    assert boss.vocabulary_ == vocabulary_desired

    arr_actual = boss.fit(X, y).transform(X).toarray()
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0)
    assert boss.vocabulary_ == vocabulary_desired


def test_accurate_results_floats():
    """Test that the actual results are the expected ones."""
    boss = BOSS(
        word_size=4, n_bins=3, window_size=0.5, window_step=0.5,
        anova=False, drop_sum=False, norm_mean=False, norm_std=False,
        strategy='quantile', alphabet=None, numerosity_reduction=True
    )

    X_windowed = X.reshape(8, 2, 100).reshape(16, 100)
    sfa = SymbolicFourierApproximation(
        n_coefs=4, drop_sum=False, anova=False, norm_mean=False,
        norm_std=False, n_bins=3, strategy='quantile', alphabet=None
    )
    y_repeated = np.repeat(y, 2)
    X_sfa = sfa.fit_transform(X_windowed, y_repeated)
    X_word = np.asarray([''.join(X_sfa[i]) for i in range(16)])
    X_word = X_word.reshape(8, 2)
    not_equal = np.c_[X_word[:, 1:] != X_word[:, :-1], np.full(8, True)]
    X_bow = np.asarray([' '.join(X_word[i, not_equal[i]]) for i in range(8)])

    vectorizer = CountVectorizer()
    arr_desired = vectorizer.fit_transform(X_bow).toarray()
    vocabulary_desired = {value: key for key, value in
                          vectorizer.vocabulary_.items()}

    arr_actual_1 = boss.fit_transform(X, None).toarray()
    np.testing.assert_allclose(arr_actual_1, arr_desired, atol=1e-5, rtol=0)
    assert boss.vocabulary_ == vocabulary_desired

    arr_actual_2 = boss.fit(X, None).transform(X).toarray()
    np.testing.assert_allclose(arr_actual_2, arr_desired, atol=1e-5, rtol=0)
    assert boss.vocabulary_ == vocabulary_desired



"""Code for Learning Time-Series Shapelets algorithm."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

from itertools import chain
from math import ceil
from numba import njit, prange
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.cluster import KMeans
from sklearn.exceptions import ConvergenceWarning
from sklearn.multiclass import OneVsOneClassifier, OneVsRestClassifier
from sklearn.multiclass import _ovr_decision_function
from sklearn.preprocessing import LabelBinarizer, LabelEncoder
from sklearn.utils import compute_class_weight, check_array
from sklearn.utils.validation import (
    check_is_fitted, check_random_state, check_X_y, _check_sample_weight)
from sklearn.utils.multiclass import check_classification_targets
import warnings
from ..utils.utils import _windowed_view
from ..base import UnivariateClassifierMixin


@njit(fastmath=True)
def _expit(x):
    """Compute the expit (logistic) function."""
    return 1 / (1 + np.exp(-x))


@njit(fastmath=True)
def _xlogy(x, y):
    """Compute the x * log(y) function."""
    return x * np.log(y)


@njit(fastmath=True)
def _softmin(arr, alpha):
    """Derive the soft-minimum of an array."""
    maximum = np.max(alpha * arr)
    exp = np.exp(alpha * arr - maximum)
    num = np.sum(arr * exp)
    den = np.sum(exp)
    return num / den


@njit(fastmath=True)
def _softmin_grad(arr, alpha):
    """Derive the gradient of the softmin function."""
    minimum = _softmin(arr, alpha)
    maximum = np.max(alpha * arr)
    exp = np.exp(alpha * arr - maximum)
    num = exp * (1 + alpha * (arr - minimum))
    den = np.sum(exp)
    return num / den


@njit(fastmath=True)
def _softmax(X, n_samples, n_classes):
    """Derive the softmax of a 2D-array."""
    maximum = np.empty((n_samples, 1))
    for i in prange(n_samples):
        maximum[i, 0] = np.max(X[i])
    exp = np.exp(X - maximum)
    sum_ = np.empty((n_samples, 1))
    for i in prange(n_samples):
        sum_[i, 0] = np.sum(exp[i])
    return exp / sum_


@njit(fastmath=True)
def _derive_shapelet_distances(X, shapelet, alpha):
    """Derive the distance between a shapelet and all the time series."""
    n_samples, n_windows, _ = X.shape

    # Derive all squared distances
    mean = np.empty((n_samples, n_windows))
    for i in prange(n_samples):
        for j in prange(n_windows):
            mean[i, j] = np.mean((X[i, j] - shapelet) ** 2)

    # Derive the soft minimum of all the distances
    dist = np.empty(n_samples)
    for i in prange(n_samples):
        dist[i] = _softmin(mean[i], alpha)

    return dist


@njit()
def _derive_all_squared_distances(
    X, n_samples, n_timestamps, shapelets, lengths, alpha
):
    """Derive the squared distances between all shapelets and time series."""
    distances = []  # save the distances in a list

    for i in prange(len(lengths)):
        window_size = lengths[i][0]
        X_window = _windowed_view(X, n_samples, n_timestamps,
                                  window_size, window_step=1)
        for j in prange(shapelets[i].shape[0]):
            dist = _derive_shapelet_distances(X_window, shapelets[i][j], alpha)
            distances.append(dist)

    return distances


@njit()
def _reshape_list_shapelets(shapelets, lengths):
    """Reshape shapelets from a 1D-array to a list of 2D-arrays."""
    shapelets_reshaped = []
    start = 0
    for length in lengths:
        n_shapelets = length.size
        length_ = length[0]
        end = start + n_shapelets * length_
        shapelets_reshaped.append(shapelets[start: end].reshape(-1, length_))
        start = end
    return shapelets_reshaped


@njit()
def _reshape_array_shapelets(shapelets, lengths):
    """Reshape shapelets from a tuple of 2D-arrays to a 1D-array."""
    lengths_concatenated = np.concatenate(lengths)
    size = np.sum(lengths_concatenated)
    shapelets_reshaped = np.empty(size)
    start = 0
    for i in range(len(shapelets)):
        end = start + np.sum(lengths[i])
        shapelets_reshaped[start:end] = np.ravel(shapelets[i])
        start = end
    return shapelets_reshaped


def _loss(X, y, n_classes, weights, shapelets, lengths, alpha, penalty, C,
          fit_intercept, intercept_scaling, sample_weight):
    """Compute the objective function."""
    n_samples, n_timestamps = X.shape

    # Derive distances between shapelets and time series
    distances = _derive_all_squared_distances(
        X, n_samples, n_timestamps, shapelets, lengths, alpha)
    distances = np.asarray(distances).T

    # Add intercept
    if fit_intercept:
        distances = np.c_[np.ones(n_samples) * intercept_scaling, distances]

    # Derive probabilities and cross-entropy loss
    if weights.ndim == 1:
        proba = _expit(distances @ weights)
        proba = np.clip(proba, 1e-8, 1 - 1e-8)
        loss_value = - np.mean(
            sample_weight * (_xlogy(y, proba) + _xlogy(1 - y, 1 - proba)))
    else:
        proba = _softmax(distances @ weights, n_samples, n_classes)
        proba = np.clip(proba, 1e-8, 1 - 1e-8)
        loss_value = - np.mean(
            sample_weight * np.sum(y * np.log(proba), axis=1))

    # Add regularization
    if penalty == 'l2':
        loss_value += (1 / C) * np.square(weights).sum()
    elif penalty == 'l1':
        loss_value += (1 / C) * np.abs(weights).sum()

    return loss_value


def _grad_weights(X, y, n_classes, weights, shapelets, lengths, alpha, penalty,
                  C, fit_intercept, intercept_scaling, sample_weight):
    """Compute the gradient of the loss with regards to the weights."""
    n_samples, n_timestamps = X.shape

    # Derive distances between shapelets and time series
    distances = _derive_all_squared_distances(
        X, n_samples, n_timestamps, shapelets, lengths, alpha)
    distances = np.asarray(distances).T

    # Add intercept
    if fit_intercept:
        distances = np.c_[np.ones(n_samples) * intercept_scaling, distances]

    # Derive probabilities and binary cross-entropy loss
    if weights.ndim == 1:
        proba = _expit(distances @ weights)
        proba = np.clip(proba, 1e-8, 1 - 1e-8)
        gradients = ((proba - y)[:, None] *
                     distances *
                     sample_weight).mean(axis=0)
    else:
        proba = _softmax(distances @ weights, n_samples, n_classes)
        proba = np.clip(proba, 1e-8, 1 - 1e-8)
        gradients = ((proba - y)[:, None, :] *
                     (distances * sample_weight)[:, :, None]).mean(axis=0)

    if penalty == 'l2':
        gradients += (2 / C) * weights
    elif penalty == 'l1':
        gradients += (1 / C) * np.sign(weights)

    return gradients


@njit()
def _compute_shapelet_grad(
    X, n_samples, n_timestamps, weights, shapelets, lengths, alpha,
    proba_minus_y, weight_idx, sample_weight,
):

    gradients = []

    for i in range(len(lengths)):
        X_window = _windowed_view(X, n_samples, n_timestamps,
                                  window_size=lengths[i][0], window_step=1)
        n_windows = X_window.shape[1]
        size = shapelets[i][0].size
        for j in range(shapelets[i].shape[0]):
            # Get the current shapelet
            shapelet = shapelets[i][j]

            # Derive the difference and the distance
            diff = shapelet - X_window
            dist = np.empty((n_samples, n_windows))
            for k in prange(n_samples):
                for m in prange(n_windows):
                    dist[k, m] = np.mean(diff[k, m] ** 2)

            # Derive the softmin gradient
            softmin_gradient = np.empty((n_samples, n_windows))
            for k in prange(n_samples):
                softmin_gradient[k] = _softmin_grad(dist[k], alpha)

            # Normalize the difference for actual gradient
            diff *= (2 / size)

            # Compute the gradient
            grad = np.empty(size)
            if weights.ndim == 1:
                for k in prange(size):
                    grad[k] = np.mean(
                        np.sum(diff[:, :, k] * softmin_gradient, axis=1) *
                        weights[weight_idx] *
                        proba_minus_y *
                        sample_weight
                    )
            else:
                for k in prange(size):
                    grad[k] = np.mean(
                        np.sum(diff[:, :, k] * softmin_gradient, axis=1) *
                        np.sum(weights[weight_idx] * proba_minus_y, axis=1) *
                        sample_weight
                    )
            gradients.append(grad)

            # Update the weight index
            weight_idx += 1

    return gradients


def _grad_shapelets(X, y, n_classes, weights, shapelets, lengths, alpha,
                    penalty, C, fit_intercept, intercept_scaling,
                    sample_weight):
    """Compute the gradient of the loss with regards to the shapelets."""
    n_samples, n_timestamps = X.shape

    # Derive distances between shapelets and time series
    distances = _derive_all_squared_distances(
        X, n_samples, n_timestamps, shapelets, lengths, alpha)
    distances = np.asarray(distances).T

    # Add intercept
    if fit_intercept:
        distances = np.c_[np.ones(n_samples) * intercept_scaling, distances]
        weight_idx = 1
    else:
        weight_idx = 0

    # Derive probabilities and cross-entropy loss
    if weights.ndim == 1:
        proba = _expit(distances @ weights)
        proba = np.clip(proba, 1e-8, 1 - 1e-8)
    else:
        proba = _softmax(distances @ weights, n_samples, n_classes)
        proba = np.clip(proba, 1e-8, 1 - 1e-8)

    # Reshape some arrays
    if weights.ndim == 1:
        proba_minus_y = (proba - y)[:, None]
    else:
        proba_minus_y = proba - y

    # Compute the gradients
    gradients = _compute_shapelet_grad(
        X, n_samples, n_timestamps, weights, shapelets, lengths,
        alpha, proba_minus_y, weight_idx, sample_weight
    )
    gradients = np.concatenate(gradients)

    return gradients


class CrossEntropyLearningShapelets(BaseEstimator, UnivariateClassifierMixin):
    """Learning Shapelets algorithm with cross-entropy loss.

    Parameters
    ----------
    n_shapelets_per_size : int or float (default = 0.2)
        Number of shapelets per size. If float, it represents
        a fraction of the number of timestamps and the number
        of shapelets per size is equal to
        ``ceil(n_shapelets_per_size * n_timestamps)``.

    min_shapelet_length : int or float (default = 0.1)
        Minimum length of the shapelets. If float, it represents
        a fraction of the number of timestamps and the minimum
        length of the shapelets per size is equal to
        ``ceil(min_shapelet_length * n_timestamps)``.

    shapelet_scale : int (default = 3)
        The different scales for the lengths of the shapelets.
        The lengths of the shapelets are equal to
        ``min_shapelet_length * np.arange(1, shapelet_scale + 1)``.
        The total number of shapelets (and features)
        is equal to ``n_shapelets_per_size * shapelet_scale``.

    penalty : 'l1' or 'l2' (default = 'l2')
        Used to specify the norm used in the penalization.

    tol : float (default = 1e-3)
        Relative tolerance for stopping criterion.

    C : float (default = 1000)
        Inverse of regularization strength. It must be a positive float.
        Smaller values specify stronger regularization.

    learning_rate : float (default = 1.)
        Learning rate for gradient descent optimization. It must be a positive
        float. Note that the learning rate will be automatically decreased
        if the loss function is not decreasing.

    max_iter : int (default = 1000)
        Maximum number of iterations for gradient descent algorithm.

    alpha : float (default = -100)
        Scaling term in the softmin function. The lower, the more precised
        the soft minimum will be. Default value should be good for
        standardized time series.

    fit_intercept : bool (default = True)
        Specifies if a constant (a.k.a. bias or intercept) should be
        added to the decision function.

    intercept_scaling : float (default = 1.)
        Scaling of the intercept. Only used if ``fit_intercept=True``.

    class_weight : dict, None or 'balanced' (default = None)
        Weights associated with classes in the form ``{class_label: weight}``.
        If not given, all classes are supposed to have unit weight.
        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input data
        as ``n_samples / (n_classes * np.bincount(y))``.

    verbose : int (default = 0)
        Controls the verbosity. It must be a non-negative integer.
        If positive, loss at each iteration is printed.

    random_state : None, int or RandomState instance (default = None)
        The seed of the pseudo random number generator to use when shuffling
        the data. If int, random_state is the seed used by the random number
        generator. If RandomState instance, random_state is the random number
        generator. If None, the random number generator is the RandomState
        instance used by `np.random`.

    Attributes
    ----------
    classes_ : array, shape = (n_classes,)
        An array of class labels known to the classifier.

    shapelets_ : array, shape = (n_shapelets,)
        Learned shapelets.

    coef_ : array, shape = (1, n_shapelets) or (n_classes, n_shapelets)
        Coefficients for each shapelet in the decision function.

    intercept_ : array, shape = (1,) or (n_classes,)
        Intercepts (a.k.a. biases) added to the decision function.
        If ``fit_intercept=False``, the intercepts are set to zero.

    n_iter_ : int
        Actual number of iterations.

    References
    ----------
    .. [1] J. Grabocka, N. Schilling, M. Wistuba and L. Schmidt-Thieme,
           "Learning Time-Series Shapelets". International Conference on Data
           Mining, 14, 392-401 (2014).

    """

    def __init__(self, n_shapelets_per_size=0.2, min_shapelet_length=0.1,
                 shapelet_scale=3, penalty='l2', tol=0.001, C=1000,
                 learning_rate=1., max_iter=1000, alpha=-100,
                 fit_intercept=True, intercept_scaling=1.,
                 class_weight=None, verbose=0, random_state=None):
        self.n_shapelets_per_size = n_shapelets_per_size
        self.min_shapelet_length = min_shapelet_length
        self.shapelet_scale = shapelet_scale
        self.penalty = penalty
        self.tol = tol
        self.C = C
        self.learning_rate = learning_rate
        self.max_iter = max_iter
        self.alpha = alpha
        self.fit_intercept = fit_intercept
        self.intercept_scaling = intercept_scaling
        self.class_weight = class_weight
        self.verbose = verbose
        self.random_state = random_state

    def fit(self, X, y, sample_weight=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Training vector.

        y : array-like, shape = (n_samples,)
            Class labels for each data sample.

        sample_weight : None or array-like, shape = (n_samples,) (default = None)
            Array of weights that are assigned to individual samples.
            If not provided, then each sample is given unit weight.

        Returns
        -------
        self : object

        """  # noqa: E501
        X, y = check_X_y(X, y)
        n_samples, n_timestamps = X.shape
        check_classification_targets(y)
        le = LabelEncoder().fit(y)
        y_ind = le.transform(y)
        self.classes_ = le.classes_
        n_classes = len(le.classes_)

        (n_shapelets_per_size, min_shapelet_length, sample_weight,
         rng) = self._check_params(X, y, y_ind, le.classes_, sample_weight)

        if n_classes > 2:
            y_ind = LabelBinarizer().fit_transform(y)

        # Shapelet initialization
        window_sizes = np.arange(
            min_shapelet_length,
            min_shapelet_length * (self.shapelet_scale + 1),
            min_shapelet_length
        )

        n_shapelets_per_cluster = n_timestamps - window_sizes + 1
        if np.any(n_shapelets_per_size > n_shapelets_per_cluster):
            raise ValueError("'n_shapelets_per_size' is too high given "
                             "'min_shapelet_length' and 'shapelet_scale'.")

        shapelets = []
        lengths = []
        for window_size in window_sizes:
            X_window = _windowed_view(
                X, n_samples, n_timestamps, window_size, window_step=1)
            X_window = X_window.reshape(-1, window_size)
            kmeans = KMeans(n_clusters=n_shapelets_per_size, random_state=rng)
            kmeans.fit(X_window)
            shapelets.append(kmeans.cluster_centers_)
            lengths.append(np.full(n_shapelets_per_size, window_size))
        shapelets = tuple(shapelets)
        lengths = tuple(lengths)

        # Weight initialization
        n_shapelets = n_shapelets_per_size * self.shapelet_scale
        if n_classes == 2:
            if self.fit_intercept:
                weights = rng.randn(n_shapelets + 1) / 100
            else:
                weights = rng.randn(n_shapelets) / 100
        else:
            if self.fit_intercept:
                weights = rng.randn(n_shapelets + 1, n_classes) / 100
            else:
                weights = rng.randn(n_shapelets, n_classes) / 100

        # Gradient descent
        learning_rate = self.learning_rate
        losses = []
        iteration = 0
        loss_iteration = _loss(
            X, y_ind, n_classes, weights, shapelets, lengths, self.alpha,
            self.penalty, self.C, self.fit_intercept, self.intercept_scaling,
            sample_weight
        )
        if self.verbose:
            print('Iteration {0}: loss = {1:0.6f}'.format(0, loss_iteration))
        losses.append(loss_iteration)
        for iteration in range(1, self.max_iter + 1):

            # Update weights
            gradient_weights = _grad_weights(
                X, y_ind, n_classes, weights, shapelets, lengths, self.alpha,
                self.penalty, self.C, self.fit_intercept,
                self.intercept_scaling, sample_weight
            )
            weights -= learning_rate * gradient_weights

            # Update shapelets
            gradient_shapelets = _grad_shapelets(
                X, y_ind, n_classes, weights, shapelets, lengths, self.alpha,
                self.penalty, self.C, self.fit_intercept,
                self.intercept_scaling, sample_weight
            )
            shapelets_array = _reshape_array_shapelets(shapelets, lengths)
            shapelets_array -= learning_rate * gradient_shapelets
            shapelets = tuple(
                _reshape_list_shapelets(shapelets_array, lengths))

            # Compute current loss
            loss_iteration = _loss(
                X, y_ind, n_classes, weights, shapelets, lengths, self.alpha,
                self.penalty, self.C, self.fit_intercept,
                self.intercept_scaling, sample_weight
            )

            # If loss is increasing, decrease the learning rate
            if losses[-1] < loss_iteration:
                while losses[-1] < loss_iteration:
                    # Go back to previous state
                    weights += learning_rate * gradient_weights
                    shapelets_array = _reshape_array_shapelets(
                        shapelets, lengths)
                    shapelets_array += learning_rate * gradient_shapelets
                    shapelets = tuple(
                        _reshape_list_shapelets(shapelets_array, lengths))

                    # Update learning  rate
                    learning_rate /= 5

                    # Recompute shapelet gradient
                    weights -= learning_rate * gradient_weights
                    gradient_shapelets = _grad_shapelets(
                        X, y_ind, n_classes, weights, shapelets, lengths,
                        self.alpha, self.penalty, self.C, self.fit_intercept,
                        self.intercept_scaling, sample_weight
                    )
                    shapelets_array = _reshape_array_shapelets(
                        shapelets, lengths)
                    shapelets_array -= learning_rate * gradient_shapelets
                    shapelets = tuple(
                        _reshape_list_shapelets(shapelets_array, lengths))

                    loss_iteration = _loss(
                        X, y_ind, n_classes, weights, shapelets, lengths,
                        self.alpha, self.penalty, self.C, self.fit_intercept,
                        self.intercept_scaling, sample_weight
                    )
            if self.verbose:
                print('Iteration {0}: loss = {1:0.6f}'
                      .format(iteration, loss_iteration))
            losses.append(loss_iteration)

            # Stopping criterion
            if abs(losses[-2] - losses[-1]) < self.tol * losses[-1]:
                break

        if iteration == self.max_iter:
            warnings.warn('Maximum number of iterations reached without '
                          'converging. Increase the maximum number of '
                          'iterations.', ConvergenceWarning)

        # Save results in attributes
        self._shapelets = shapelets
        self._lengths = lengths
        self.shapelets_ = [list(shapelet) for shapelet in shapelets]
        self.shapelets_ = np.asarray(
            list(chain.from_iterable(self.shapelets_)), dtype='object')
        if n_classes == 2:
            if self.fit_intercept:
                self.intercept_ = np.array([weights[0]])
                self.coef_ = weights[1:].reshape(1, -1)
            else:
                self.intercept_ = np.array([0])
                self.coef_ = weights.reshape(1, -1)
        else:
            if self.fit_intercept:
                self.intercept_ = weights[0]
                self.coef_ = weights[1:].T
            else:
                self.intercept_ = np.zeros(weights.shape[1])
                self.coef_ = weights.T
        self.n_iter_ = iteration

    def decision_function(self, X):
        """Decision function scores.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_timestamps)
            Test samples.

        Returns
        -------
        T : array-like of shape (n_samples,) or (n_samples, n_classes)
            Decision function scores for each sample for each class in the
            model, where classes are ordered as they are in ``self.classes_``.

        """
        check_is_fitted(self, ['shapelets_', 'coef_', 'intercept_', 'n_iter_'])

        X = check_array(X)
        n_samples, n_timestamps = X.shape

        # Derive distances between shapelets and time series
        distances = _derive_all_squared_distances(
            X, n_samples, n_timestamps, self._shapelets,
            self._lengths, self.alpha
        )
        distances = np.asarray(distances).T

        # Add intercept
        if self.fit_intercept:
            distances = np.c_[np.ones(n_samples) * self.intercept_scaling,
                              distances]

        # Derive decision function
        if self.fit_intercept:
            if len(self.classes_) == 2:
                weights = np.r_[self.intercept_, np.squeeze(self.coef_)]
            else:
                weights = np.r_[self.intercept_.reshape(1, -1), self.coef_.T]
        else:
            weights = self.coef_.T
        X_new = np.squeeze(distances @ weights)

        return X_new

    def predict_proba(self, X):
        """Probability estimates.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_timestamps)
            Test samples.

        Returns
        -------
        T : array-like of shape (n_samples, n_classes)
            Probability of the samples for each class in the model,
            where classes are ordered as they are in ``self.classes_``.

        """
        X_new = self.decision_function(X)
        n_samples = X_new.shape[0]
        if len(self.classes_) == 2:
            proba = _expit(X_new)
            X_proba = np.c_[1 - proba, proba]
        else:
            X_proba = _softmax(X_new, n_samples, len(self.classes_))
        return X_proba

    def predict(self, X):
        """Predict the class labels for the provided data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Test samples.

        Returns
        -------
        y_pred : array-like, shape = (n_samples,)
            Class labels for each data sample.

        """
        if len(self.classes_) == 2:
            y_pred = (self.decision_function(X) > 0.).astype('int64')
        else:
            y_pred = self.decision_function(X).argmax(axis=1)
        return self.classes_[y_pred]

    def _check_params(self, X, y, y_ind, classes, sample_weight):
        """Parameter check"""
        n_samples, n_timestamps = X.shape

        if not isinstance(self.n_shapelets_per_size,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'n_shapelets_per_size' must be an integer or a "
                            "float (got {})."
                            .format(self.n_shapelets_per_size))
        if isinstance(self.n_shapelets_per_size, (int, np.integer)):
            if not 1 <= self.n_shapelets_per_size <= n_timestamps:
                raise ValueError(
                    "If 'n_shapelets_per_size' is an integer, it must be "
                    "greater than or equal to 1 and lower than or equal to "
                    "n_timestamps (got {0}).".format(self.n_shapelets_per_size)
                )
            n_shapelets_per_size = self.n_shapelets_per_size
        else:
            if not (0 < self.n_shapelets_per_size <= 1.):
                raise ValueError(
                    "If 'n_shapelets_per_size' is a float, it must be greater "
                    "than 0 and lower than or equal to 1 (got {0})."
                    .format(self.n_shapelets_per_size)
                )
            n_shapelets_per_size = ceil(
                self.n_shapelets_per_size * n_timestamps)

        if not isinstance(self.min_shapelet_length,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'min_shapelet_length' must be an integer or a "
                            "float (got {}).".format(self.min_shapelet_length))
        if isinstance(self.min_shapelet_length, (int, np.integer)):
            if not 1 <= self.min_shapelet_length <= n_timestamps:
                raise ValueError(
                    "If 'min_shapelet_length' is an integer, it must be "
                    "greater than or equal to 1 and lower than or equal to "
                    "n_timestamps (got {0}).".format(self.min_shapelet_length)
                )
            min_shapelet_length = self.min_shapelet_length
        else:
            if not (0 < self.min_shapelet_length <= 1.):
                raise ValueError(
                    "If 'min_shapelet_length' is a float, it must be greater "
                    "than 0 and lower than or equal to 1 (got {0})."
                    .format(self.min_shapelet_length)
                )
            min_shapelet_length = ceil(self.min_shapelet_length * n_timestamps)

        if not (isinstance(self.shapelet_scale, (int, np.integer)) and
                self.shapelet_scale > 0):
            raise ValueError("'shapelet_scale' must be a positive integer "
                             "(got {}).".format(self.shapelet_scale))

        if self.shapelet_scale * min_shapelet_length > n_timestamps:
            raise ValueError(
                "'shapelet_scale' and 'min_shapelet_length' must be "
                "such that shapelet_scale * min_shapelet_length is "
                "smaller than or equal to n_timestamps."
            )

        if self.penalty not in ('l1', 'l2'):
            raise ValueError("'penalty' must be either 'l2' or 'l1' "
                             "(got {}).".format(self.penalty))

        if not (isinstance(self.C, (int, np.integer, float, np.floating)) and
                self.C > 0):
            raise ValueError("'C' must be a positive float (got {})."
                             .format(self.C))

        if not (isinstance(self.tol, (int, np.integer, float, np.floating)) and
                self.tol > 0):
            raise ValueError("'tol' must be a positive float (got {})."
                             .format(self.tol))

        if not (isinstance(self.learning_rate,
                           (int, np.integer, float, np.floating)) and
                self.learning_rate > 0):
            raise ValueError("'learning_rate' must be a positive float "
                             "(got {}).".format(self.learning_rate))

        if not (isinstance(self.max_iter, (int, np.integer)) and
                self.max_iter >= 0):
            raise ValueError("'max_iter' must be a non-negative integer "
                             "(got {}).".format(self.max_iter))

        if not (isinstance(self.alpha, (int, np.integer, float, np.floating))
                and self.alpha < 0):
            raise ValueError("'alpha' must be a negative float (got {})."
                             .format(self.alpha))

        if not isinstance(self.intercept_scaling,
                          (int, np.integer, float, np.floating)):
            raise ValueError("'intercept_scaling' must be a float (got {})."
                             .format(self.intercept_scaling))

        class_weight_balanced = (isinstance(self.class_weight, str) and
                                 self.class_weight == 'balanced')
        if not (self.class_weight is None or
                class_weight_balanced or
                isinstance(self.class_weight, dict)):
            raise ValueError("'class_weight' must be None, a dictionary "
                             " or 'balanced' (got {})."
                             .format(self.class_weight))
        class_weight = compute_class_weight(
            self.class_weight, classes=classes, y=y)

        sample_weight = _check_sample_weight(sample_weight, X, dtype='float64')
        sample_weight *= class_weight[y_ind]
        sample_weight = sample_weight.reshape(-1, 1)

        rng = check_random_state(self.random_state)

        if not (isinstance(self.verbose, (int, np.integer)) and
                self.verbose >= 0):
            raise ValueError("'verbose' must be a non-negative integer "
                             "(got {}).".format(self.verbose))

        return n_shapelets_per_size, min_shapelet_length, sample_weight, rng


class LearningShapelets(BaseEstimator, UnivariateClassifierMixin):
    """Learning Shapelets algorithm.

    This estimator consists of two steps: computing the distances between the
    shapelets and the time series, then computing a logistic regression using
    these distances as features. This algorithm learns the shapelets as well as
    the coefficients of the logistic regression.

    Parameters
    ----------
    n_shapelets_per_size : int or float (default = 0.2)
        Number of shapelets per size. If float, it represents
        a fraction of the number of timestamps and the number
        of shapelets per size is equal to
        ``ceil(n_shapelets_per_size * n_timestamps)``.

    min_shapelet_length : int or float (default = 0.1)
        Minimum length of the shapelets. If float, it represents
        a fraction of the number of timestamps and the minimum
        length of the shapelets per size is equal to
        ``ceil(min_shapelet_length * n_timestamps)``.

    shapelet_scale : int (default = 3)
        The different scales for the lengths of the shapelets.
        The lengths of the shapelets are equal to
        ``min_shapelet_length * np.arange(1, shapelet_scale + 1)``.
        The total number of shapelets (and features)
        is equal to ``n_shapelets_per_size * shapelet_scale``.

    penalty : 'l1' or 'l2' (default = 'l2')
        Used to specify the norm used in the penalization.

    tol : float (default = 1e-3)
        Tolerance for stopping criterion.

    C : float (default = 1000)
        Inverse of regularization strength. It must be a positive float.
        Smaller values specify stronger regularization.

    learning_rate : float (default = 1.)
        Learning rate for gradient descent optimization. It must be a positive
        float. Note that the learning rate will be automatically decreased
        if the loss function is not decreasing.

    max_iter : int (default = 1000)
        Maximum number of iterations for gradient descent algorithm.

    multi_class : {'multinomial', 'ovr', 'ovo'} (default = 'multinomial')
        Strategy for multiclass classification.
        'multinomial' stands for multinomial cross-entropy loss.
        'ovr' stands for one-vs-rest strategy.
        'ovo' stands for one-vs-one strategy.
        Ignored if the classification task is binary.

    alpha : float (default = -100)
        Scaling term in the softmin function. The lower, the more precised
        the soft minimum will be. Default value should be good for
        standardized time series.

    fit_intercept : bool (default = True)
        Specifies if a constant (a.k.a. bias or intercept) should be
        added to the decision function.

    intercept_scaling : float (default = 1.)
        Scaling of the intercept. Only used if ``fit_intercept=True``.

    class_weight : dict, None or 'balanced' (default = None)
        Weights associated with classes in the form ``{class_label: weight}``.
        If not given, all classes are supposed to have unit weight.
        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input data
        as ``n_samples / (n_classes * np.bincount(y))``.

    verbose : int (default = 0)
        Controls the verbosity. It must be a non-negative integer.
        If positive, loss at each iteration is printed.

    random_state : None, int or RandomState instance (default = None)
        The seed of the pseudo random number generator to use when shuffling
        the data. If int, random_state is the seed used by the random number
        generator. If RandomState instance, random_state is the random number
        generator. If None, the random number generator is the RandomState
        instance used by `np.random`.

    n_jobs : None or int (default = None)
        The number of jobs to use for the computation. Only used if
        ``multi_class`` is 'ovr' or 'ovo'.

    Attributes
    ----------
    classes_ : array, shape = (n_classes,)
        An array of class labels known to the classifier.

    shapelets_ : array shape = (n_tasks, n_shapelets)
        Learned shapelets. Each element of this array is a learned
        shapelet.

    coef_ : array, shape = (n_tasks, n_shapelets) or (n_classes, n_shapelets)
        Coefficients for each shapelet in the decision function.

    intercept_ : array, shape = (n_tasks,) or (n_classes,)
        Intercepts (a.k.a. biases) added to the decision function.
        If ``fit_intercept=False``, the intercepts are set to zero.

    n_iter_ : array, shape = (n_tasks,)
        Actual number of iterations.

    Notes
    -----
    The number of tasks (n_tasks) depends on the value of ``multi_class``
    and the number of classes. If there are two classes, the number of
    tasks is equal to 1. If there are more than two classes, the number
    of tasks is equal to:

        - 1 if ``multi_class='multinomial'``
        - n_classes if ``multi_class='ovr'``
        - n_classes * (n_classes - 1) / 2 if ``multi_class='ovo'``

    References
    ----------
    .. [1] J. Grabocka, N. Schilling, M. Wistuba and L. Schmidt-Thieme,
           "Learning Time-Series Shapelets". International Conference on Data
           Mining, 14, 392-401 (2014).

    Examples
    --------
    >>> from pyts.classification import LearningShapelets
    >>> X = [[1, 2, 2, 1, 2, 3, 2],
    ...      [0, 2, 0, 2, 0, 2, 3],
    ...      [0, 1, 2, 2, 1, 2, 2]]
    >>> y = [0, 1, 0]
    >>> clf = LearningShapelets(random_state=42, tol=0.01)
    >>> clf.fit(X, y)
    LearningShapelets(...)
    >>> clf.coef_.shape
    (1, 6)

    """

    def __init__(self, n_shapelets_per_size=0.2, min_shapelet_length=0.1,
                 shapelet_scale=3, penalty='l2', tol=0.001, C=1000,
                 learning_rate=1., max_iter=1000, multi_class='multinomial',
                 alpha=-100, fit_intercept=True, intercept_scaling=1.,
                 class_weight=None, verbose=0, random_state=None, n_jobs=None):
        self.n_shapelets_per_size = n_shapelets_per_size
        self.min_shapelet_length = min_shapelet_length
        self.shapelet_scale = shapelet_scale
        self.penalty = penalty
        self.tol = tol
        self.C = C
        self.learning_rate = learning_rate
        self.max_iter = max_iter
        self.multi_class = multi_class
        self.alpha = alpha
        self.fit_intercept = fit_intercept
        self.intercept_scaling = intercept_scaling
        self.class_weight = class_weight
        self.verbose = verbose
        self.random_state = random_state
        self.n_jobs = n_jobs

    def fit(self, X, y, sample_weight=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Training vector.

        y : array-like, shape = (n_samples,)
            Class labels for each data sample.

        sample_weight : None or array-like, shape = (n_samples,) (default = None)
            Array of weights that are assigned to individual samples.
            If not provided, then each sample is given unit weight.

        Returns
        -------
        self : object

        """  # noqa: E501
        X, y = check_X_y(X, y)
        n_classes = len(LabelEncoder().fit(y).classes_)
        multi_class = self._check_params(n_classes)
        params = self.get_params()
        params.pop('n_jobs', None)
        params.pop('multi_class')
        clf = CrossEntropyLearningShapelets(**params)
        if multi_class == 'ovr':
            clf = OneVsRestClassifier(clf, n_jobs=self.n_jobs)
        elif multi_class == 'ovo':
            clf = OneVsOneClassifier(clf, n_jobs=self.n_jobs)
        clf.fit(X, y)

        self.classes_ = clf.classes_
        self._multi_class = multi_class

        if multi_class in ('ovr', 'ovo'):
            self._estimators = clf.estimators_
            self.shapelets_ = np.array(
                [est.shapelets_ for est in clf.estimators_])
            self.coef_ = np.squeeze(np.asarray(
                [est.coef_ for est in clf.estimators_]))
            self.intercept_ = np.squeeze(np.array(
                [est.intercept_ for est in clf.estimators_]))
            self.n_iter_ = np.array(
                [est.n_iter_ for est in clf.estimators_])
        else:
            self._clf = clf
            self.shapelets_ = np.array([clf.shapelets_])
            self.coef_ = clf.coef_
            self.intercept_ = clf.intercept_
            self.n_iter_ = np.array([clf.n_iter_])
        return self

    def decision_function(self, X):
        """Decision function scores.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Test samples.

        Returns
        -------
        T : array, shape = (n_samples,) or (n_samples, n_classes)
            Decision function scores for each sample for each class in the
            model, where classes are ordered as they are in ``self.classes_``.

        """
        X = check_array(X)

        if self._multi_class == 'ovr':
            X_new = np.empty((X.shape[0], self.classes_.size))
            for i, estimator in enumerate(self._estimators):
                X_new[:, i] = estimator.decision_function(X)
        elif self._multi_class == 'ovo':
            predictions = np.vstack(
                [est.predict(X) for est in self._estimators]).T
            confidences = np.vstack(
                [est.decision_function(X) for est in self._estimators]).T
            X_new = _ovr_decision_function(
                predictions, confidences, len(self.classes_))
        else:
            X_new = self._clf.decision_function(X)

        return X_new

    def predict_proba(self, X):
        """Probability estimates.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Test samples.

        Returns
        -------
        T : array, shape = (n_samples, n_classes)
            Probability of the samples for each class in the model,
            where classes are ordered as they are in ``self.classes_``.

        """
        X_new = self.decision_function(X)
        n_samples = X_new.shape[0]
        if self._multi_class == 'binary':
            proba = _expit(X_new)
            return np.c_[1 - proba, proba]
        elif self._multi_class == 'multinomial':
            proba = _softmax(X_new, n_samples, len(self.classes_))
        else:
            # OvR normalization, like LibLinear's predict_probability
            proba = _expit(X_new)
            proba /= proba.sum(axis=1).reshape((proba.shape[0], -1))
        return proba

    def predict(self, X):
        """Predict the class labels for the provided data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Test samples.

        Returns
        -------
        y_pred : array-like, shape = (n_samples,)
            Class labels for each data sample.

        """
        X_new = self.decision_function(X)
        if X_new.ndim == 2:
            y_pred = X_new.argmax(axis=1)
        else:
            y_pred = (X_new > 0).astype('int64')
        return self.classes_[y_pred]

    def _check_params(self, n_classes):
        if self.multi_class not in ('multinomial', 'ovr', 'ovo'):
            raise ValueError(
                "'multi_class' must be either 'multinomial', "
                "'ovr' or 'ovo' (got {}).".format(self.multi_class)
            )
        multi_class = 'binary' if n_classes == 2 else self.multi_class

        class_weight_dict = isinstance(self.class_weight, dict)
        if multi_class in ('ovr', 'ovo') and class_weight_dict:
            raise ValueError("'class_weight' must be None or 'balanced' if "
                             "'multi_class' is either 'ovr' or 'ovo'.")

        n_jobs_int = (isinstance(self.n_jobs, (int, np.integer)) and
                      self.n_jobs != 0)
        if not (self.n_jobs is None or n_jobs_int):
            raise ValueError("'n_jobs' must be None or an integer not equal "
                             "to zero (got {}).".format(self.n_jobs))

        return multi_class



"""Testing for utility tools."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re

from pyts.multivariate.utils import check_3d_array

n_samples, n_features, n_timestamps = 40, 3, 30
rng = np.random.RandomState(42)
X = rng.randn(n_samples, n_features, n_timestamps)


@pytest.mark.parametrize(
    'X, err_msg',
    [(X[0, 0], "X must be 3-dimensional (got 1)."),
     (X[0], "X must be 3-dimensional (got 2).")]
)
def test_3d_input(X, err_msg):
    """Test input data validation."""
    with pytest.raises(ValueError, match=re.escape(err_msg)):
        check_3d_array(X)



"""Code for imputers."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from scipy.interpolate import interp1d
from sklearn.base import BaseEstimator
from sklearn.impute import MissingIndicator
from sklearn.utils.validation import check_array
from ..base import UnivariateTransformerMixin


class InterpolationImputer(BaseEstimator, UnivariateTransformerMixin):
    """Impute missing values using interpolation.

    Parameters
    ----------
    missing_values : None, np.nan, integer or float (default = np.nan)
        The placeholder for the missing values. All occurrences of
        `missing_values` will be imputed. If an integer or a float,
        the input data must not contain NaN or infinity values.

    strategy : str or int (default = 'linear')
        Specifies the kind of interpolation as a string
        ('linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic',
        'previous', 'next', where 'zero', 'slinear', 'quadratic' and 'cubic'
        refer to a spline interpolation of zeroth, first, second or third
        order; 'previous' and 'next' simply return the previous or next value
        of the point) or as an integer specifying the order of the spline
        interpolator to use. Default is 'linear'.

    Examples
    --------
    >>> import numpy as np
    >>> from pyts.preprocessing import InterpolationImputer
    >>> X = [[1, None, 3, 4], [8, None, 4, None]]
    >>> imputer = InterpolationImputer()
    >>> imputer.transform(X)
    array([[1., 2., 3., 4.],
           [8., 6., 4., 2.]])

    """

    def __init__(self, missing_values=np.nan, strategy='linear'):
        self.missing_values = missing_values
        self.strategy = strategy

    def fit(self, X=None, y=None):
        """Pass.

        Parameters
        ----------
        X
            Ignored

        y
            Ignored

        Returns
        -------
        self : object

        """
        return self

    def transform(self, X):
        """Perform imputation using interpolation.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data with missing values.

        Returns
        -------
        X_new : array-like, shape = (n_samples, n_timestamps)
            Data without missing values.

        """
        missing_values, force_all_finite = self._check_params()
        X = check_array(X, dtype='float64', force_all_finite=force_all_finite)
        n_samples, n_timestamps = X.shape

        indicator = MissingIndicator(
            missing_values=missing_values, features='all', sparse=False,
        )
        non_missing_idx = ~(indicator.fit_transform(X))
        x_new = np.arange(n_timestamps)
        X_imputed = np.asarray(
            [self._impute_one_sample(X[i], non_missing_idx[i], x_new)
             for i in range(n_samples)]
        )
        return X_imputed

    def _check_params(self):
        if self.missing_values is None:
            missing_values = np.nan
            force_all_finite = 'allow-nan'
        elif isinstance(self.missing_values,
                        (int, np.integer, float, np.floating)):
            if np.isinf(self.missing_values):
                raise ValueError("'missing_values' cannot be infinity.")
            elif np.isnan(self.missing_values):
                force_all_finite = 'allow-nan'
                missing_values = np.nan
            else:
                force_all_finite = True
                missing_values = self.missing_values
        else:
            raise ValueError(
                "'missing_values' must be an integer, a float, None or "
                "np.nan (got {0!s})".format(self.missing_values)
            )
        strategy_str_values = [
            'linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic',
            'previous', 'next'
        ]
        if not ((isinstance(self.strategy, (int, np.integer)))
                or (self.strategy in strategy_str_values)):
            raise ValueError(
                "'strategy' must be an integer or one of 'linear', 'nearest', "
                "'zero', 'slinear', 'quadratic', 'cubic', 'previous', 'next' "
                "(got {0})".format(self.strategy)
            )
        return missing_values, force_all_finite

    def _impute_one_sample(self, x, non_missing_idx, x_new):
        idx = x_new[non_missing_idx]
        f = interp1d(idx, x[non_missing_idx], kind=self.strategy,
                     copy=True, fill_value='extrapolate', assume_sorted=True)
        return f(x_new)



"""The :mod:`pyts.classification` module includes classification algorithms."""

from .bossvs import BOSSVS
from .learning_shapelets import LearningShapelets
from .knn import KNeighborsClassifier
from .saxvsm import SAXVSM
from .time_series_forest import TimeSeriesForest
from .tsbf import TSBF

__all__ = ['BOSSVS', 'KNeighborsClassifier', 'LearningShapelets', 'SAXVSM',
           'TimeSeriesForest', 'TSBF']



"""Code for transformers."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

from sklearn.base import BaseEstimator
from sklearn.preprocessing import PowerTransformer as SklearnPowerTransformer
from sklearn.preprocessing import (QuantileTransformer as
                                   SklearnQuantileTransformer)
from sklearn.utils.validation import check_array
from ..base import UnivariateTransformerMixin


class PowerTransformer(BaseEstimator, UnivariateTransformerMixin):
    """Apply a power transform sample-wise to make data more Gaussian-like.

    Power transforms are a family of parametric, monotonic transformations
    that are applied to make data more Gaussian-like. This is useful for
    modeling issues related to heteroscedasticity (non-constant variance),
    or other situations where normality is desired.

    Currently, PowerTransformer supports the Box-Cox transform and the
    Yeo-Johnson transform. The optimal parameter for stabilizing variance and
    minimizing skewness is estimated through maximum likelihood.

    Box-Cox requires input data to be strictly positive, while Yeo-Johnson
    supports both positive or negative data.

    By default, zero-mean, unit-variance normalization is applied to the
    transformed data.

    Parameters
    ----------
    method : 'yeo-johnson' or 'box-cox' (default = 'yeo-johnson')
        The power transform method. Available methods are:

        - 'yeo-johnson' [1]_, works with positive and negative values
        - 'box-cox' [2]_, only works with strictly positive values

    standardize : boolean (default = True)
        Set to True to apply zero-mean, unit-variance normalization to the
        transformed output.

    Notes
    -----
    NaNs are treated as missing values: disregarded in ``fit``, and maintained
    in ``transform``.

    References
    ----------
    .. [1] I.K. Yeo and R.A. Johnson, "A new family of power transformations to
           improve normality or symmetry." Biometrika, 87(4), pp.954-959,
           (2000).
    .. [2] G.E.P. Box and D.R. Cox, "An Analysis of Transformations", Journal
           of the Royal Statistical Society B, 26, 211-252 (1964).

    Examples
    --------
    >>> import numpy as np
    >>> from pyts.preprocessing import PowerTransformer
    >>> X = [[1, 3, 4], [2, 2, 5]]
    >>> pt = PowerTransformer()
    >>> print(pt.transform(X))
    [[-1.316...  0.209...  1.106...]
     [-0.707... -0.707...  1.414...]]

    """

    def __init__(self, method='yeo-johnson', standardize=True):
        self.method = method
        self.standardize = standardize

    def fit(self, X=None, y=None):
        """Pass.

        Parameters
        ----------
        X
            Ignored

        y
            Ignored

        Returns
        -------
        self : object

        """
        return self

    def transform(self, X):
        """Transform the data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data to transform.

        Returns
        -------
        X_new : array-like, shape = (n_samples, n_timestamps)
            Transformed data.

        """
        X = check_array(X, dtype='float64', force_all_finite='allow-nan')
        transformer = SklearnPowerTransformer(
            method=self.method, standardize=self.standardize
        )
        X_new = transformer.fit_transform(X.T).T
        return X_new


class QuantileTransformer(BaseEstimator, UnivariateTransformerMixin):
    """Transform samples using quantiles information.

    This method transforms the samples to follow a uniform or a normal
    distribution. Therefore, for a given sample, this transformation tends
    to spread out the most frequent values. It also reduces the impact of
    (marginal) outliers: this is therefore a robust preprocessing scheme.
    The transformation is applied on each sample independently.

    The cumulative distribution function of a feature is used to project the
    original values. Note that this transform is non-linear.

    Parameters
    ----------
    n_quantiles : int, optional (default = 1000)
        Number of quantiles to be computed. It corresponds to the number
        of landmarks used to discretize the cumulative distribution function.

    output_distribution : 'uniform' or 'normal' (default = 'uniform')
        Marginal distribution for the transformed data. The choices are
        'uniform' (default) or 'normal'.

    subsample : int, optional (default = 1e5)
        Maximum number of timestamps used to estimate the quantiles for
        computational efficiency.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by np.random. Note that this is used by subsampling and smoothing
        noise.

    Examples
    --------
    >>> from pyts.datasets import load_gunpoint
    >>> from pyts.preprocessing import QuantileTransformer
    >>> X, _, _, _  = load_gunpoint(return_X_y=True)
    >>> qt = QuantileTransformer(n_quantiles=10)
    >>> qt.transform(X)
    array([...])

    """

    def __init__(self, n_quantiles=1000, output_distribution='uniform',
                 subsample=int(1e5), random_state=None):
        self.n_quantiles = n_quantiles
        self.output_distribution = output_distribution
        self.subsample = subsample
        self.random_state = random_state

    def fit(self, X=None, y=None):
        """Pass.

        Parameters
        ----------
        X
            Ignored

        y
            Ignored

        Returns
        -------
        self : object

        """
        return self

    def transform(self, X):
        """Transform the data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data to transform.

        Returns
        -------
        X_new : array-like, shape = (n_samples, n_timestamps)
            Transformed data.

        """
        X = check_array(X, dtype='float64')
        transformer = SklearnQuantileTransformer(
            n_quantiles=self.n_quantiles,
            output_distribution=self.output_distribution,
            subsample=self.subsample,
            random_state=self.random_state
        )
        X_new = transformer.fit_transform(X.T).T
        return X_new



"""Testing for Learning Time-Series Shapelets algorithm."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re

from sklearn.exceptions import ConvergenceWarning

from pyts.classification.learning_shapelets import (
    CrossEntropyLearningShapelets
)
from pyts.classification import LearningShapelets
from pyts.classification.learning_shapelets import (
    _expit, _xlogy, _softmin, _softmin_grad, _softmax,
    _derive_shapelet_distances, _derive_all_squared_distances,
    _reshape_list_shapelets, _reshape_array_shapelets,
    _loss, _grad_weights, _grad_shapelets
)


X_multi = np.arange(20, dtype='float64').reshape(4, 5)
y_multi = np.arange(4)
X_bin = X_multi[:2]
y_bin = y_multi[:2]
n_classes = 2
shapelets = (np.array([[4., 6.]]), np.array([[0., 1., 3., 3.]]))
lengths = (np.array([2]), np.array([4]))


@pytest.mark.parametrize(
    'x, arr_desired',
    [(0, 0.5),
     (np.arange(10), 1 / (1 + np.exp(- np.arange(10)))),
     (np.log(np.arange(1, 10)), 1 / (1 + 1 / np.arange(1, 10))),
     (np.log(1 / np.arange(1, 10)), 1 / (1 + np.arange(1, 10)))]
)
def test_expit(x, arr_desired):
    arr_actual = _expit(x)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0)


@pytest.mark.parametrize(
    'x, y, arr_desired',
    [(0, 3, 0),
     (3, 1, 0),
     (np.arange(6), np.arange(6, 12), np.arange(6) * np.log(np.arange(6, 12))),
     (np.array([0, 1]), np.array([2, 1]), [0, 0])]
)
def test_xlogy(x, y, arr_desired):
    arr_actual = _xlogy(x, y)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0)


@pytest.mark.parametrize(
    'arr, alpha, min_desired',
    [(np.arange(100), -50, 0),
     (np.arange(-100, 100), -50, -100),
     (np.arange(0, 10000, 100), -500, 0),
     (np.linspace(2, 3, 50), -500, 2)]
)
def test_softmin(arr, alpha, min_desired):
    min_actual = _softmin(arr, alpha)
    np.testing.assert_allclose(min_actual, min_desired, atol=1e-5, rtol=0)


@pytest.mark.parametrize(
    'arr, alpha, grad_desired',
    [(np.arange(100), -50, [1] + [0] * 99),
     (np.arange(-100, 100), -50, [1] + [0] * 199),
     (np.arange(0, 10000, 100), -500, [1] + [0] * 99),
     (np.linspace(2, 3, 50), -5000, [1] + [0] * 49)]
)
def test_softmin_grad(arr, alpha, grad_desired):
    grad_actual = _softmin_grad(arr, alpha)
    np.testing.assert_allclose(grad_actual, grad_desired, atol=1e-5, rtol=0)


@pytest.mark.parametrize(
    'arr, min_desired',
    [([np.log(np.arange(1, 5))], [np.arange(0.1, 0.5, 0.1)]),
     ([np.log(np.arange(10, 14))], [np.arange(10, 14) / 46]),
     ([np.log(np.arange(1, 10))], [np.arange(1, 10) / 45])]
)
def test_softmax(arr, min_desired):
    arr = np.asarray(arr)
    n_samples, n_classes = arr.shape
    min_actual = _softmax(arr, n_samples, n_classes)
    np.testing.assert_allclose(min_actual, min_desired, atol=1e-5, rtol=0)


@pytest.mark.parametrize(
    'X, shapelet, arr_desired',
    [(np.ones((5, 3, 2)), np.ones(2), np.zeros(5)),
     (np.ones((5, 3, 2)), np.zeros(2), np.ones(5)),
     (2 * np.ones((5, 3, 2)), np.zeros(2), 4 * np.ones(5)),
     (np.arange(30).reshape(2, 3, 5), np.arange(-2, 3), [4, 289]),
     (np.arange(30).reshape(2, 3, 5), np.zeros(5), [6, 291])]
)
def test_derive_shapelet_distances(X, shapelet, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = _derive_shapelet_distances(X, shapelet, alpha=-100)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'X': np.arange(6, dtype='float64').reshape(2, 3), 'n_samples': 2,
       'n_timestamps': 3, 'shapelets': (np.array([[1, 2], [2, 1]]),),
       'lengths': (np.array([2, 2]),)},
      [[0, 4], [1, 5]]),

     ({'X': np.arange(8, dtype='float64').reshape(2, 4), 'n_samples': 2,
       'n_timestamps': 4, 'lengths': (np.array([2]), np.array([3])),
       'shapelets': (np.array([[1, 2]]), np.array([[2, 1, 2]]))},
      [[0, 9], [1, 12]])
     ]
)
def test_derive_all_squared_distances(params, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = _derive_all_squared_distances(**params, alpha=-100)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'shapelets, lengths, res_desired',
    [(np.arange(13), (np.array([3, 3, 3]), np.array([2, 2])),
      [np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]]),
       np.array([[9, 10], [11, 12]])]),

     (np.arange(10), (np.array([5, 5]),),
      [np.arange(10).reshape(2, 5)])]
)
def test_reshape_list_shapelets(shapelets, lengths, res_desired):
    """Test that the actual results are the expected ones."""
    res_actual = _reshape_list_shapelets(shapelets, lengths)
    assert len(res_actual) == len(res_desired)
    for arr_actual, arr_desired in zip(res_actual, res_desired):
        np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'shapelets, lengths, arr_desired',
    [((np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]]),
       np.array([[9, 10], [11, 12]])),
      (np.array([3, 3, 3]), np.array([2, 2])), np.arange(13)),

     ((np.arange(10).reshape(2, 5),),
      (np.array([5, 5]),), np.arange(10))]
)
def test_reshape_array_shapelets(shapelets, lengths, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = _reshape_array_shapelets(shapelets, lengths)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'shapelets, lengths',
    [(np.arange(13), (np.array([3, 3, 3]), np.array([2, 2]))),
     (np.arange(10), (np.array([5, 5]),))]
)
def test_reshape_list_shapelets_inverse(shapelets, lengths):
    """Test that the actual results are the expected ones."""
    shapelets_tuple = tuple(_reshape_list_shapelets(shapelets, lengths))
    shapelets_array = _reshape_array_shapelets(shapelets_tuple, lengths)
    np.testing.assert_allclose(shapelets, shapelets_array, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'shapelets, lengths',
    [((np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]]),
       np.array([[9, 10], [11, 12]])),
      (np.array([3, 3, 3]), np.array([2, 2]))),
     ((np.arange(10).reshape(2, 5),), (np.array([5, 5]),))]
)
def test_reshape_array_shapelets_inverse(shapelets, lengths):
    """Test that the actual results are the expected ones."""
    shapelets_array = _reshape_array_shapelets(shapelets, lengths)
    shapelets_tuple = tuple(_reshape_list_shapelets(shapelets_array, lengths))
    assert len(shapelets) == len(shapelets_tuple)
    for arr_actual, arr_desired in zip(shapelets, shapelets_tuple):
        np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, loss_desired',
    [({'weights': np.zeros(2), 'fit_intercept': False, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([1., 1.])}, np.log(2)),
     ({'weights': np.zeros(3), 'fit_intercept': True, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([1., 1.])}, np.log(2)),
     ({'weights': np.zeros((3, 2)), 'fit_intercept': True, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([2., 2.])}, 2 * np.log(2)),
     ({'weights': np.array([-2, 1]), 'fit_intercept': False, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([1., 1.])},
      0.00861448 / 2 + 5),
     ({'weights': np.array([1, -2, 1]), 'fit_intercept': True, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([1., 1.])},
      0.02324546 / 2 + 6),
     ({'weights': np.array([-2, 1]), 'fit_intercept': False, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([3., 1.])},
      0.00861448 * 1.5 + 5),
     ({'weights': np.array([-2, 1]), 'fit_intercept': False, 'penalty': 'l2',
       'C': 10, 'sample_weight': np.array([3., 1.])},
      0.00861448 * 1.5 + 0.5),
     ({'weights': np.array([-2, 1]), 'fit_intercept': False, 'penalty': 'l1',
       'C': 1, 'sample_weight': np.array([3., 1.])},
      0.00861448 * 1.5 + 3),
     ({'weights': np.array([-2, 1]), 'fit_intercept': False, 'penalty': 'l1',
       'C': 10, 'sample_weight': np.array([3., 1.])},
      0.00861448 * 1.5 + 0.3),
     ({'weights': np.array([[-2, -2], [1, 1]]), 'fit_intercept': False,
       'penalty': 'l2', 'C': 1, 'sample_weight': np.array([1., 1.])},
      np.log(2) + 10),
     ({'weights': np.array([[3, 3], [-2, -2], [1, 1]]), 'fit_intercept': True,
       'penalty': 'l2', 'C': 1, 'sample_weight': np.array([1., 1.])},
      np.log(2) + 28)]
)
def test_loss(params, loss_desired):
    """Test that the actual loss is equal to the expected loss."""
    loss_actual = _loss(X_bin, y_bin, n_classes, shapelets=shapelets,
                        lengths=lengths, alpha=-100, intercept_scaling=1,
                        **params)
    np.testing.assert_allclose(loss_actual, loss_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'weights': np.zeros(2), 'fit_intercept': False, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([1., 1.])}, [0.5, -5.625]),
     ({'weights': np.zeros(3), 'fit_intercept': True, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([1., 1.])}, [0, 0.5, -5.625]),
     ({'weights': np.zeros(3), 'fit_intercept': True, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([2., 2.])}, [0, 1, -11.25]),
     ({'weights': np.zeros((3, 2)), 'fit_intercept': True, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([1., 1.])},
      [[0.5, -0.5], [0.75, -0.75], [5.75, -5.75]]),
     ({'weights': np.array([-2, 1]), 'fit_intercept': False, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([1., 1.])},
      [-3.98927814, 2.00107218]),
     ({'weights': np.array([1, -2, 1]), 'fit_intercept': True, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([1., 1.])},
      [2.01148868, -3.97127829, 2.00287217]),
     ({'weights': np.array([-2, 1]), 'fit_intercept': False, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([3., 1.])},
      [-3.96783443, 2.00321655]),
     ({'weights': np.array([-2, 1]), 'fit_intercept': False, 'penalty': 'l2',
       'C': 10, 'sample_weight': np.array([3., 1.])},
      [-0.36783443, 0.20321655]),
     ({'weights': np.array([-2, 1]), 'fit_intercept': False, 'penalty': 'l1',
       'C': 1, 'sample_weight': np.array([3., 1.])},
      [-0.96783443, 1.00321655]),
     ({'weights': np.array([-2, 1]), 'fit_intercept': False, 'penalty': 'l1',
       'C': 10, 'sample_weight': np.array([3., 1.])},
      [-0.06783443, 0.10321655]),
     ({'weights': np.array([[-2, -2], [1, 1]]), 'fit_intercept': False,
       'penalty': 'l2', 'C': 1, 'sample_weight': np.array([1., 1.])},
      [[-3.25, -4.75], [7.75, -3.75]]),
     ({'weights': np.array([[3, 3], [-2, -2], [1, 1]]), 'fit_intercept': True,
       'penalty': 'l2', 'C': 1, 'sample_weight': np.array([1., 1.])},
      [[6.5, 5.5], [-3.25, -4.75], [7.75, -3.75]])]
)
def test_grad_weights(params, arr_desired):
    """Test that the actual gradient is the expected gradient."""
    params['sample_weight'] = params['sample_weight'].reshape(-1, 1)
    arr_actual = _grad_weights(
        X_bin, y_bin, n_classes, shapelets=shapelets, lengths=lengths,
        alpha=-100, intercept_scaling=1, **params
    )
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'weights': np.zeros(2), 'fit_intercept': False, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([1., 1.])}, np.zeros(6)),
     ({'weights': np.zeros(3), 'fit_intercept': True, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([1., 1.])}, np.zeros(6)),
     ({'weights': np.zeros(3), 'fit_intercept': True, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([2., 2.])}, np.zeros(6)),
     ({'weights': np.zeros((3, 2)), 'fit_intercept': True, 'penalty': 'l2',
       'C': 1, 'sample_weight': np.array([1., 1.])}, np.zeros(6)),
     ({'weights': np.zeros(3), 'fit_intercept': True, 'penalty': 'l1',
       'C': 1, 'sample_weight': np.array([2., 2.])}, np.zeros(6)),
     ({'weights': np.zeros(3), 'fit_intercept': True, 'penalty': 'l2',
       'C': 10, 'sample_weight': np.array([2., 2.])}, np.zeros(6)),
     ({'weights': np.array([3., 0., 0.]), 'fit_intercept': True,
       'penalty': 'l2', 'C': 10, 'sample_weight': np.array([1., 1.])},
      np.zeros(6))]
)
def test_grad_shapelets(params, arr_desired):
    """Test that the actual gradient is the expected gradient."""
    params['sample_weight'] = params['sample_weight'].reshape(-1, 1)
    arr_actual = _grad_shapelets(
        X_bin, y_bin, n_classes, shapelets=shapelets, lengths=lengths,
        alpha=-100, intercept_scaling=1, **params
    )
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'n_shapelets_per_size': 'str'}, TypeError,
      "'n_shapelets_per_size' must be an integer or a float (got str)."),

     ({'n_shapelets_per_size': 8}, ValueError,
      "If 'n_shapelets_per_size' is an integer, it must be "
      "greater than or equal to 1 and lower than or equal to "
      "n_timestamps (got 8)."),

     ({'n_shapelets_per_size': -0.5}, ValueError,
      "If 'n_shapelets_per_size' is a float, it must be greater "
      "than 0 and lower than or equal to 1 (got -0.5)."),

     ({'min_shapelet_length': [6]}, TypeError,
      "'min_shapelet_length' must be an integer or a float (got [6])."),

     ({'min_shapelet_length': 6}, ValueError,
      "If 'min_shapelet_length' is an integer, it must be "
      "greater than or equal to 1 and lower than or equal to "
      "n_timestamps (got 6)."),

     ({'min_shapelet_length': 2.0}, ValueError,
      "If 'min_shapelet_length' is a float, it must be greater "
      "than 0 and lower than or equal to 1 (got 2.0)."),

     ({'shapelet_scale': 'yolo'}, ValueError,
      "'shapelet_scale' must be a positive integer (got yolo)."),

     ({'shapelet_scale': 0}, ValueError,
      "'shapelet_scale' must be a positive integer (got 0)."),

     ({'shapelet_scale': 12}, ValueError,
      "'shapelet_scale' and 'min_shapelet_length' must be "
      "such that shapelet_scale * min_shapelet_length is "
      "smaller than or equal to n_timestamps."),

     ({'penalty': 'l3'}, ValueError,
      "'penalty' must be either 'l2' or 'l1' (got l3)."),

     ({'C': 'l3'}, ValueError,
      "'C' must be a positive float (got l3)."),

     ({'C': -1}, ValueError,
      "'C' must be a positive float (got -1)."),

     ({'tol': 'l4'}, ValueError,
      "'tol' must be a positive float (got l4)."),

     ({'tol': 0}, ValueError,
      "'tol' must be a positive float (got 0)."),

     ({'learning_rate': 'l5'}, ValueError,
      "'learning_rate' must be a positive float (got l5)."),

     ({'learning_rate': 0}, ValueError,
      "'learning_rate' must be a positive float (got 0)."),

     ({'max_iter': '200'}, ValueError,
      "'max_iter' must be a non-negative integer (got 200)."),

     ({'max_iter': 200.}, ValueError,
      "'max_iter' must be a non-negative integer (got 200.0)."),

     ({'alpha': '200'}, ValueError,
      "'alpha' must be a negative float (got 200)."),

     ({'alpha': 200}, ValueError,
      "'alpha' must be a negative float (got 200)."),

     ({'intercept_scaling': '300'}, ValueError,
      "'intercept_scaling' must be a float (got 300)."),

     ({'class_weight': [3]}, ValueError,
      "'class_weight' must be None, a dictionary  or 'balanced' (got [3])."),

     ({'verbose': '200'}, ValueError,
      "'verbose' must be a non-negative integer (got 200)."),

     ({'verbose': -1}, ValueError,
      "'verbose' must be a non-negative integer (got -1)."),

     ({'n_shapelets_per_size': 5}, ValueError,
      "'n_shapelets_per_size' is too high given "
      "'min_shapelet_length' and 'shapelet_scale'.")]
)
def test_parameter_check_cross_entropy(params, error, err_msg):
    """Test parameter validation."""
    clf = CrossEntropyLearningShapelets(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        clf.fit(X_bin, y_bin)


@pytest.mark.parametrize('X, y', [(X_bin, y_bin), (X_multi, y_multi)])
@pytest.mark.parametrize('max_iter', [0, 1, 2])
def test_convergence_warning(X, y, max_iter):
    """Test that the ConvergenceWarning is raised."""
    clf = CrossEntropyLearningShapelets(learning_rate=1e-3, tol=1e-9,
                                        random_state=42, max_iter=max_iter)
    msg = ('Maximum number of iterations reached without converging. Increase '
           'the maximum number of iterations.')
    with pytest.warns(ConvergenceWarning, match=msg):
        clf.fit(X, y)


@pytest.mark.filterwarnings("ignore:Maximum number of iterations")
@pytest.mark.parametrize('X, y', [(X_bin, y_bin), (X_multi, y_multi)])
@pytest.mark.parametrize(
    'params, n_shapelets_desired',
    [({'verbose': 1, 'learning_rate': 100, 'max_iter': 1}, 3),
     ({'min_shapelet_length': 1, 'fit_intercept': False}, 3),
     ({'n_shapelets_per_size': 3}, 9),
     ({'shapelet_scale': 5, 'tol': 10, 'max_iter': 2}, 5),
     ({'n_shapelets_per_size': 2, 'shapelet_scale': 3}, 6)]
)
def test_shapes_cross_entropy(X, y, params, n_shapelets_desired):
    """Test that the attributes and returned arrays have the expected shapes"""
    n_samples = X.shape[0]
    if 'max_iter' not in params.keys():
        params['max_iter'] = 0
    clf = CrossEntropyLearningShapelets(**params)

    # Training phase
    clf.fit(X, y)
    n_classes = 1 if len(clf.classes_) == 2 else len(clf.classes_)
    np.testing.assert_array_equal(clf.classes_, y)
    assert clf.shapelets_.shape == (n_shapelets_desired,)
    assert clf.coef_.shape == (n_classes, n_shapelets_desired)
    assert clf.intercept_.shape == (n_classes,)
    assert clf.n_iter_ <= clf.max_iter

    # Test phase
    X_new = clf.decision_function(X)
    y_proba = clf.predict_proba(X)
    y_pred = clf.predict(X)
    if n_classes == 1:
        assert X_new.shape == (n_samples,)
        assert y_proba.shape == (n_samples, 2)
    else:
        assert X_new.shape == (n_samples, n_classes)
        assert y_proba.shape == (n_samples, n_classes)
    assert y_pred.shape == (n_samples,)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'multi_class': 'yolo'}, ValueError,
      "'multi_class' must be either 'multinomial', 'ovr' or 'ovo' "
      "(got yolo)."),

     ({'class_weight': {0: 2, 1: 3}, 'multi_class': 'ovr'}, ValueError,
      "'class_weight' must be None or 'balanced' if 'multi_class' is "
      "either 'ovr' or 'ovo'."),

     ({'class_weight': {0: 2, 1: 3}, 'multi_class': 'ovo'}, ValueError,
      "'class_weight' must be None or 'balanced' if 'multi_class' is "
      "either 'ovr' or 'ovo'."),

     ({'n_jobs': 0}, ValueError,
      "'n_jobs' must be None or an integer not equal to zero (got 0)."),

     ({'n_jobs': 'oops'}, ValueError,
      "'n_jobs' must be None or an integer not equal to zero (got oops).")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    clf = LearningShapelets(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        clf.fit(X_multi, y_multi)


@pytest.mark.filterwarnings("ignore:Maximum number of iterations")
@pytest.mark.parametrize('X, y', [(X_bin, y_bin), (X_multi, y_multi)])
@pytest.mark.parametrize(
    'params, n_shapelets_desired, n_tasks',
    [({'verbose': 1, 'learning_rate': 100, 'max_iter': 1}, 3, 4),
     ({'multi_class': 'ovr'}, 3, 4),
     ({'multi_class': 'ovo'}, 3, 6),
     ({'fit_intercept': False}, 3, 4),
     ({'n_shapelets_per_size': 3}, 9, 4),
     ({'shapelet_scale': 5}, 5, 4),
     ({'n_shapelets_per_size': 2, 'shapelet_scale': 3}, 6, 4)]
)
def test_shapes(X, y, params, n_shapelets_desired, n_tasks):
    """Test that the attributes and returned arrays have the expected shapes"""
    n_samples = X.shape[0]
    if 'max_iter' not in params.keys():
        params['max_iter'] = 0
    clf = LearningShapelets(**params)

    # Training phase
    clf.fit(X, y)
    n_classes = len(clf.classes_) if len(clf.classes_) > 2 else 1
    np.testing.assert_array_equal(clf.classes_, y)
    if clf._multi_class in ('binary', 'multinomial'):
        assert clf.shapelets_.shape == (1, n_shapelets_desired)
        assert clf.coef_.shape == (n_classes, n_shapelets_desired)
        assert clf.intercept_.shape == (n_classes,)
        assert clf.n_iter_.shape == (1,)
    else:
        assert clf.shapelets_.shape == (n_tasks, n_shapelets_desired)
        assert clf.coef_.shape == (n_tasks, n_shapelets_desired)
        assert clf.intercept_.shape == (n_tasks,)
        assert clf.n_iter_.shape == (n_tasks,)

    # Test phase
    X_new = clf.decision_function(X)
    y_proba = clf.predict_proba(X)
    y_pred = clf.predict(X)
    if clf._multi_class == 'binary':
        assert X_new.shape == (n_samples,)
        assert y_proba.shape == (n_samples, 2)
    else:
        assert X_new.shape == (n_samples, n_classes)
        assert y_proba.shape == (n_samples, n_classes)
    assert y_pred.shape == (n_samples,)



"""Testing for datatset generation."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.datasets import make_cylinder_bell_funnel


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'n_samples': '3'}, TypeError, "'n_samples' must be an integer."),

     ({'n_samples': -1}, ValueError,
      "'n_samples' must be a positive integer."),

     ({'weights': [1]}, ValueError,
      "'weights' must be None or a list with 2 or 3 elements (got 1)."),

     ({'weights': [1, 2, 3, 4]}, ValueError,
      "'weights' must be None or a list with 2 or 3 elements (got 4)."),

     ({'weights': [0.5, 1.5]}, ValueError,
      "'sum(weights)' cannot be larger than 1 if len(weights) == 2 (got 2.0)")]
)
def test_parameter_check_make_cylinder_bell_funnel(params, error, err_msg):
    """Test parameter validation."""
    with pytest.raises(error, match=re.escape(err_msg)):
        make_cylinder_bell_funnel(**params)


@pytest.mark.parametrize(
    'params, class_balance_desired',
    [({'n_samples': 9}, [3, 3, 3]),
     ({'n_samples': 90}, [30, 30, 30]),
     ({'n_samples': 10, 'weights': [0., 0.5, 0.5]}, [0, 5, 5]),
     ({'n_samples': 10, 'weights': [0., 1., 1.]}, [0, 10, 10]),
     ({'n_samples': 10, 'weights': [0., 1., 0.]}, [0, 10]),
     ({'n_samples': 10, 'weights': np.array([0., 0.5, 0.5])}, [0, 5, 5]),
     ({'n_samples': 10, 'weights': (0., 0.5, 0.5)}, [0, 5, 5]),
     ({'n_samples': 10, 'weights': (0.5, 0.5, 0.)}, [5, 5]),
     ({'n_samples': 10, 'weights': (0.5, 0.5)}, [5, 5]),
     ({'n_samples': 10, 'weights': (0.8, 0.2)}, [8, 2])]
)
def test_class_balance_make_cylinder_bell_funnel(params,
                                                 class_balance_desired):
    """Test that the class balance is the expected one."""
    X, y = make_cylinder_bell_funnel(**params)
    class_balance_actual = np.bincount(y)
    np.testing.assert_array_equal(class_balance_actual, class_balance_desired)


@pytest.mark.parametrize(
    'params',
    [{},
     {'return_params': True},
     {'n_samples': 100, 'return_params': True}])
def test_return_params_make_cylinder_bell_funnel(params):
    """Test the return objects."""
    res = make_cylinder_bell_funnel(**params)
    assert isinstance(res[0], np.ndarray)
    assert isinstance(res[1], np.ndarray)
    if 'return_params' in params.keys() and params['return_params']:
        parameters = res[2]
        assert isinstance(parameters, dict)
        assert isinstance(parameters['a'], (int, np.integer))
        assert isinstance(parameters['b'], (int, np.integer))
        assert isinstance(parameters['eta'], (float, np.floating))
        assert isinstance(parameters['epsilon'], np.ndarray)
        if 'n_samples' in params:
            assert parameters['epsilon'].shape == (params['n_samples'], 128)
        else:
            assert parameters['epsilon'].shape == (30, 128)


@pytest.mark.parametrize('params', [{}, {'shuffle': False}])
def test_shuffle_make_cylinder_bell_funnel(params):
    """Test that shuffling works as expected."""
    _, y = make_cylinder_bell_funnel(**params)
    arr_desired_no_shuffle = np.repeat(np.arange(3), 10)
    if 'shuffle' in params.keys() and (not params['shuffle']):
        np.testing.assert_array_equal(y, arr_desired_no_shuffle)
    else:
        assert not np.array_equal(y, arr_desired_no_shuffle)



"""
The :mod:`pyts.datasets` module tools for making, loading and fetching time
series datasets.
"""

from .load import (load_basic_motions, load_coffee, load_gunpoint,
                   load_pig_central_venous_pressure)
from .make import make_cylinder_bell_funnel
from .ucr import ucr_dataset_info, ucr_dataset_list, fetch_ucr_dataset
from .uea import uea_dataset_info, uea_dataset_list, fetch_uea_dataset

__all__ = ['load_basic_motions', 'load_coffee', 'load_gunpoint',
           'load_pig_central_venous_pressure', 'make_cylinder_bell_funnel',
           'ucr_dataset_info', 'ucr_dataset_list', 'fetch_ucr_dataset',
           'uea_dataset_info', 'uea_dataset_list', 'fetch_uea_dataset']



"""Testing for Multiple Coefficient Binning."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.approximation.mcb import _uniform_bins, _digitize
from pyts.approximation import MultipleCoefficientBinning


X = np.arange(30).reshape(10, 3)
y = [0, 0, 0, 1, 1, 2, 1, 0, 0, 2]


@pytest.mark.parametrize(
    'timestamp_min, timestamp_max, n_bins, arr_desired',
    [(np.arange(5), np.arange(2, 7), 2, np.arange(1, 6).reshape(5, 1)),

     ([4, 8, -2, 2], [8, 20, 2, 10], 4,
      [[5, 6, 7], [11, 14, 17], [-1, 0, 1], [4, 6, 8]]),

     ([0, 10, 10], [12, 28, 70], 6,
      [[2, 4, 6, 8, 10], [13, 16, 19, 22, 25], [20, 30, 40, 50, 60]])]
)
def test_uniform_bins(timestamp_min, timestamp_max, n_bins, arr_desired):
    """Test that the actual results are the expected ones."""
    timestamp_min = np.asarray(timestamp_min)
    timestamp_max = np.asarray(timestamp_max)
    n_timestamps = timestamp_min.size
    arr_actual = _uniform_bins(timestamp_min, timestamp_max,
                               n_timestamps, n_bins)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'X, bins, arr_desired',
    [([[0, 1], [0.2, 0.8], [0.4, 0.6], [0.6, 0.4], [0.8, 0.2], [1, 0]],
      [0.2, 0.4, 0.6, 0.8],
      [[0, 4], [0, 3], [1, 2], [2, 1], [3, 0], [4, 0]]),

     ([[0, 1], [0.4, 1.8], [0.4, 1.6], [0.6, 1.4], [0.2, 1.2], [1, 2]],
      [0.2, 0.4, 0.6, 0.8],
      [[0, 4], [1, 4], [1, 4], [2, 4], [0, 4], [4, 4]]),

     ([[0, 1], [0.4, 1.8], [0.4, 1.6], [0.6, 1.4], [0.2, 1.2], [1, 2]],
      [[0.2, 0.4, 0.6, 0.8], [0.2, 0.4, 0.6, 0.8]],
      [[0, 4], [1, 4], [1, 4], [2, 4], [0, 4], [4, 4]]),

     ([[0, 1], [0.4, 1.8], [0.4, 1.6], [0.6, 1.4], [0.2, 1.2], [1, 2]],
      [[0.2, 0.4, 0.6, 0.8], [0.2, 1.4, 1.6, 2.8]],
      [[0, 1], [1, 3], [1, 2], [2, 1], [0, 1], [4, 3]])]
)
def test_digitize(X, bins, arr_desired):
    """Test that the actual results are the expected ones."""
    X = np.asarray(X)
    bins = np.asarray(bins)
    arr_actual = _digitize(X, bins)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'n_bins': '3'}, TypeError, "'n_bins' must be an integer."),

     ({'alphabet': 'whoops'}, TypeError,
      "'alphabet' must be None, 'ordinal' or array-like with shape (n_bins,) "
      "(got {0})".format('whoops')),

     ({'n_bins': 1}, ValueError,
      "'n_bins' must be greater than or equal to 2 and lower than "
      "or equal to min(n_samples, 26) (got 1)."),

     ({'n_bins': 15}, ValueError,
      "'n_bins' must be greater than or equal to 2 and lower than "
      "or equal to min(n_samples, 26) (got 15)."),

     ({'strategy': 'whoops'}, ValueError,
      "'strategy' must be either 'uniform', 'quantile', 'normal' or 'entropy' "
      "(got {0})".format('whoops')),

     ({'alphabet': ['a', 'b', 'c']}, ValueError,
      "If 'alphabet' is array-like, its shape must be equal to (n_bins,).")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    mcb = MultipleCoefficientBinning(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        mcb.fit(X, y)


def test_constant_sample():
    """Test that a ValueError is raised with a constant sample."""
    discretizer = MultipleCoefficientBinning()
    msg_error = "At least one timestamp is constant."
    with pytest.raises(ValueError, match=msg_error):
        discretizer.fit_transform(np.ones((10, 15)))


def test_consistent_length():
    """Test that a ValueError is raised with a constant sample."""
    msg_error = (
        "The number of timestamps in X must be the same as "
        "the number of timestamps when `fit` was called "
        "({0} != {1}).".format(3, 1)
    )
    mcb = MultipleCoefficientBinning()
    mcb.fit(X, y)
    with pytest.raises(ValueError, match=re.escape(msg_error)):
        mcb.transform(X[:, :1])


def test_identical_bin_edges():
    """Test that an error is raised when two consecutive bins are equal."""
    msg_error = ("At least two consecutive quantiles are equal. "
                 "Consider trying with a smaller number of bins or "
                 "removing timestamps with low variation.")
    mcb = MultipleCoefficientBinning(n_bins=6, strategy='quantile')
    with pytest.raises(ValueError, match=msg_error):
        mcb.fit(np.r_[np.zeros((4, 2)), np.ones((4, 2))])


def test_y_none_entropy():
    """Test that an error is raised when y is None and 'entropy' is used."""
    msg_error = "y cannot be None if strategy='entropy'."
    mcb = MultipleCoefficientBinning(n_bins=2, strategy='entropy')
    with pytest.raises(ValueError, match=msg_error):
        mcb.fit(X, None)


def test_high_n_bins_entropy():
    """Test that an error is raised when 'n_bins' is too large."""
    msg_error = ("The number of bins is too high for timestamp {0}. "
                 "Consider trying with a smaller number of bins or "
                 "removing this timestamp.".format(0))
    mcb = MultipleCoefficientBinning(n_bins=9, strategy='entropy')
    with pytest.raises(ValueError, match=msg_error):
        mcb.fit(X, y)


@pytest.mark.parametrize(
    'params, X, y, arr_desired',
    [({}, X, None,
      np.asarray([['a', 'a', 'a', 'b', 'b', 'c', 'c', 'd', 'd', 'd']] * 3).T),

     ({'strategy': 'uniform'}, X, None,
      np.asarray([['a', 'a', 'a', 'b', 'b', 'c', 'c', 'd', 'd', 'd']] * 3).T),

     ({'alphabet': 'ordinal'}, X, None,
      np.asarray([[0, 0, 0, 1, 1, 2, 2, 3, 3, 3]] * 3).T),

     ({'n_bins': 10, 'strategy': 'quantile', 'alphabet': 'ordinal'}, X, None,
      np.asarray([np.arange(10)] * 3).T),

     ({'n_bins': 10, 'strategy': 'uniform', 'alphabet': 'ordinal'}, X, None,
      np.asarray([np.arange(10)] * 3).T),

     ({'n_bins': 5, 'strategy': 'quantile', 'alphabet': 'ordinal'}, X, None,
      np.asarray([np.repeat(np.arange(5), 2)] * 3).T),

     ({'n_bins': 5, 'strategy': 'uniform', 'alphabet': 'ordinal'}, X, None,
      np.asarray([np.repeat(np.arange(5), 2)] * 3).T),

     ({'n_bins': 3, 'strategy': 'normal'}, [[-0.5], [0], [0.5]], None,
      [['a'], ['b'], ['c']]),

     ({'n_bins': 6, 'strategy': 'entropy'}, X, y,
      np.asarray([['a', 'a', 'a', 'b', 'b', 'c', 'd', 'e', 'e', 'f']] * 3).T),

     ({'n_bins': 6, 'strategy': 'entropy', 'alphabet': 'ordinal'}, X, y,
      np.asarray([[0, 0, 0, 1, 1, 2, 3, 4, 4, 5]] * 3).T)]
)
def test_actual_results(params, X, y, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = MultipleCoefficientBinning(**params).fit_transform(X, y)
    np.testing.assert_array_equal(arr_actual, arr_desired)


@pytest.mark.parametrize(
    'params',
    [({}),
     ({'strategy': 'uniform'}),
     ({'alphabet': 'ordinal'}),
     ({'n_bins': 10, 'strategy': 'quantile', 'alphabet': 'ordinal'}),
     ({'n_bins': 5, 'strategy': 'quantile', 'alphabet': 'ordinal'}),
     ({'n_bins': 5, 'strategy': 'normal'}),
     ({'n_bins': 6, 'strategy': 'entropy'}),
     ({'n_bins': 6, 'strategy': 'entropy', 'alphabet': 'ordinal'})]
)
def test_fit_transform(params):
    """Test that fit and transform yield the same results as fit_transform."""
    arr_1 = MultipleCoefficientBinning(**params).fit(X, y).transform(X)
    arr_2 = MultipleCoefficientBinning(**params).fit_transform(X, y)
    np.testing.assert_array_equal(arr_1, arr_2)



"""Testing for scalers."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
from pyts.preprocessing import (StandardScaler, MinMaxScaler,
                                MaxAbsScaler, RobustScaler)


X = np.arange(10).reshape(2, 5)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'with_mean': True, 'with_std': True},
      (X - X.mean(axis=1)[:, None]) / X.std(axis=1)[:, None]),

     ({'with_mean': True, 'with_std': False}, X - X.mean(axis=1)[:, None]),

     ({'with_mean': False, 'with_std': True}, X / X.std(axis=1)[:, None]),

     ({'with_mean': False, 'with_std': False}, X)]
)
def test_actual_results_standard_scaler(params, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = StandardScaler(**params).fit_transform(X)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'sample_range': (0, 1)}, [[0, 0.25, 0.5, 0.75, 1] for _ in range(2)]),

     ({'sample_range': (-1, 1)}, [[-1, -0.5, 0, 0.5, 1] for _ in range(2)]),

     ({'sample_range': (0, 4)}, [[0, 1, 2, 3, 4] for _ in range(2)]),

     ({'sample_range': (-10, 10)}, [[-10, -5, 0, 5, 10] for _ in range(2)])]
)
def test_actual_results_min_max_scaler(params, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = MinMaxScaler(**params).fit_transform(X)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


def test_actual_results_max_abs_scaler():
    """Test that the actual results are the expected ones."""
    arr_actual = MaxAbsScaler().fit_transform(X)
    arr_desired = X / np.abs(X).max(axis=1)[:, None]
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({}, [[-1, -0.5, 0, 0.5, 1] for _ in range(2)]),

     ({'quantile_range': (0., 100.)},
      [[-0.5, -0.25, 0, 0.25, 0.5] for _ in range(2)]),

     ({'with_centering': False}, X / 2),

     ({'with_scaling': False}, [[-2, -1, 0, 1, 2] for _ in range(2)]),

     ({'with_centering': False, 'with_scaling': False}, X)]
)
def test_actual_results_robust_scaler(params, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = RobustScaler(**params).fit_transform(X)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)



"""Testing for datatset loading."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
from ..load import (load_basic_motions, load_coffee, load_gunpoint,
                    load_pig_central_venous_pressure)


@pytest.mark.parametrize(
    'function, data_train_shape, data_test_shape, target_train_shape, '
    'target_test_shape, n_classes',
    [(load_basic_motions, (40, 6, 100), (40, 6, 100), (40,), (40,), 4),
     (load_coffee, (28, 286), (28, 286), (28,), (28,), 2),
     (load_gunpoint, (50, 150), (150, 150), (50,), (150,), 2),
     (load_pig_central_venous_pressure, (104, 2000), (208, 2000), (104,),
      (208,), 52)]
)
@pytest.mark.parametrize('return_X_y', [False, True])
def test_load_functions(function, data_train_shape, data_test_shape,
                        target_train_shape, target_test_shape, n_classes,
                        return_X_y):
    """Test the loading functions."""
    res = function(return_X_y=return_X_y)
    if return_X_y:
        data_train, data_test, target_train, target_test = res
    else:
        data_train = res.data_train
        data_test = res.data_test
        target_train = res.target_train
        target_test = res.target_test
        assert isinstance(res.url, str)
        assert isinstance(res.DESCR, str)
    for data in (data_train, data_test, target_train, target_test):
        assert isinstance(data, np.ndarray)

    assert data_train.shape == data_train_shape
    assert data_test.shape == data_test_shape
    assert target_train.shape == target_train_shape
    assert target_test.shape == target_test_shape
    assert np.unique(target_train).size == n_classes
    assert np.unique(target_test).size == n_classes



"""Code for Symbolic Aggregate approXimation."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array
from ..base import UnivariateTransformerMixin
from ..preprocessing import KBinsDiscretizer


class SymbolicAggregateApproximation(BaseEstimator,
                                     UnivariateTransformerMixin):
    """Symbolic Aggregate approXimation.

    Parameters
    ----------
    n_bins : int (default = 4)
        The number of bins to produce. It must be between 2 and
        ``min(n_timestamps, 26)``.

    strategy : 'uniform', 'quantile' or 'normal' (default = 'quantile')
        Strategy used to define the widths of the bins:

        - 'uniform': All bins in each sample have identical widths
        - 'quantile': All bins in each sample have the same number of points
        - 'normal': Bin edges are quantiles from a standard normal distribution

    raise_warning : bool (default = True)
        If True, a warning is raised when the number of bins is smaller for
        at least one sample. In this case, you should consider decreasing the
        number of bins or removing these samples.

    alphabet : None, 'ordinal' or array-like, shape = (n_bins,)
        Alphabet to use. If None, the first `n_bins` letters of the Latin
        alphabet are used. If 'ordinal', integers are used.

    References
    ----------
    .. [1] J. Lin, E. Keogh, L. Wei, and S. Lonardi, "Experiencing SAX: a
           novel symbolic representation of time series". Data Mining and
           Knowledge Discovery, 15(2), 107-144 (2007).

    Examples
    --------
    >>> from pyts.approximation import SymbolicAggregateApproximation
    >>> X = [[0, 4, 2, 1, 7, 6, 3, 5],
    ...      [2, 5, 4, 5, 3, 4, 2, 3]]
    >>> transformer = SymbolicAggregateApproximation()
    >>> print(transformer.transform(X))
    [['a' 'c' 'b' 'a' 'd' 'd' 'b' 'c']
     ['a' 'd' 'c' 'd' 'b' 'c' 'a' 'b']]

    """

    def __init__(self, n_bins=4, strategy='quantile', raise_warning=True,
                 alphabet=None):
        self.n_bins = n_bins
        self.strategy = strategy
        self.raise_warning = raise_warning
        self.alphabet = alphabet

    def fit(self, X=None, y=None):
        """Pass.

        Parameters
        ----------
        X
            Ignored
        y
            Ignored

        """
        return self

    def transform(self, X):
        """Bin the data with the given alphabet.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data to transform.

        Returns
        -------
        X_new : array, shape = (n_samples, n_timestamps)
            Binned data.

        """
        X = check_array(X, dtype='float64')
        n_timestamps = X.shape[1]
        alphabet = self._check_params(n_timestamps)
        discretizer = KBinsDiscretizer(
            n_bins=self.n_bins, strategy=self.strategy,
            raise_warning=self.raise_warning
        )
        indices = discretizer.fit_transform(X)
        if isinstance(alphabet, str):
            return indices
        else:
            return alphabet[indices]

    def _check_params(self, n_timestamps):
        if not isinstance(self.n_bins, (int, np.integer)):
            raise TypeError("'n_bins' must be an integer.")
        if not 2 <= self.n_bins <= 26:
            raise ValueError(
                "'n_bins' must be greater than or equal to 2 and lower than "
                "or equal to 26 (got {0})."
                .format(self.n_bins)
            )
        if self.strategy not in ['uniform', 'quantile', 'normal']:
            raise ValueError("'strategy' must be either 'uniform', 'quantile' "
                             "or 'normal' (got {0})".format(self.strategy))
        if not ((self.alphabet is None)
                or (self.alphabet == 'ordinal')
                or (isinstance(self.alphabet, (list, tuple, np.ndarray)))):
            raise TypeError("'alphabet' must be None, 'ordinal' or array-like "
                            "with shape (n_bins,) (got {0})"
                            .format(self.alphabet))
        if self.alphabet is None:
            alphabet = np.array([chr(i) for i in range(97, 97 + self.n_bins)])
        elif self.alphabet == 'ordinal':
            alphabet = 'ordinal'
        else:
            alphabet = check_array(self.alphabet, ensure_2d=False, dtype=None)
            if alphabet.shape != (self.n_bins,):
                raise ValueError("If 'alphabet' is array-like, its shape "
                                 "must be equal to (n_bins,).")
        return alphabet



"""Testing for Dynamic Time Warping and its variants."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from math import sqrt
from numba import njit
from pyts.metrics.dtw import (
    _square, _absolute, _check_input_dtw, _blurred_path_region, _return_path,
    _return_results, cost_matrix, accumulated_cost_matrix, _check_region,
    _dtw_classic, _dtw_region, _dtw_sakoechiba, _dtw_itakura,
    _dtw_multiscale, _dtw_fast
)
from pyts.metrics import (
    dtw, itakura_parallelogram, sakoe_chiba_band, show_options
)


x = np.array([0, 1, 2])
y = np.array([2, 0, 1])

x_long = np.array([0, 1, 2, 4, 2, 0])
y_long = np.array([2, 0, 1, 3, 0, 1])

params_return = {'return_cost': True,
                 'return_accumulated': True,
                 'return_path': True}


@njit
def power_four(x, y):
    """Power four function with a njit decorator."""
    return (x - y) ** 4


@pytest.mark.parametrize(
    'x, y, scalar_desired',
    [(1, 2, 1),
     (1, 3, 4),
     (3, 1, 4),
     (5, 2, 9)]
)
def test_square(x, y, scalar_desired):
    """Test that the actual results are the expected ones."""
    assert _square(x, y) == scalar_desired


@pytest.mark.parametrize(
    'x, y, scalar_desired',
    [(1, 2, 1),
     (1, 3, 2),
     (3, 1, 2),
     (5, 2, 3)]
)
def test_absolute(x, y, scalar_desired):
    """Test that the actual results are the expected ones."""
    assert _absolute(x, y) == scalar_desired


@pytest.mark.parametrize(
    'params, err_msg',
    [({'x': x[:, None], 'y': y, 'precomputed_cost': None, 'dist': 'squared',
       'method': 'classic'}, "'x' must be a one-dimensional array."),
     ({'x': x, 'y': y[:, None], 'precomputed_cost': None, 'dist': 'squared',
       'method': 'classic'}, "'y' must be a one-dimensional array."),
     ({'x': x, 'y': y, 'precomputed_cost': None, 'dist': 'precomputed',
       'method': 'fast'}, "cannot be used with a precomputed cost")]
)
def test_check_input_dtw(params, err_msg):
    """Test parameter validation."""
    with pytest.raises(ValueError, match=re.escape(err_msg)):
        _check_input_dtw(**params)


@pytest.mark.parametrize(
    'params, err_msg',
    [({'dist': 'who'},
      "'dist' must be either 'square', 'absolute' or callable (got who)."),

     ({'dist': lambda x: x}, "Calling dist(1, 1) did not work."),

     ({'dist': lambda x, y: str(x) + str(y)},
      "Calling dist(1, 1) did not return a float or an integer."),

     ({'region': [[1, 1]]},
      "The shape of 'region' must be equal to (2, n_timestamps_1) "
      "(got (1, 2)).")]
)
def test_parameter_check_cost_matrix(params, err_msg):
    """Test parameter validation."""
    with pytest.raises(ValueError, match=re.escape(err_msg)):
        cost_matrix(x, y, **params)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({}, [[4, 0, 1], [1, 1, 0], [0, 4, 1]]),

     ({'dist': 'absolute'}, [[2, 0, 1], [1, 1, 0], [0, 2, 1]]),

     ({'dist': power_four}, [[16, 0, 1], [1, 1, 0], [0, 16, 1]]),

     ({'region': [[0, 0, 1], [2, 3, 3]]},
      [[4, 0, np.inf], [1, 1, 0], [np.inf, 4, 1]])]
)
def test_actual_results_cost_matrix(params, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = cost_matrix(x, y, **params)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, err_msg',
    [
     ({'cost_mat': np.ones((4, 4)), 'region': [[1, 1]]},
      "The shape of 'region' must be equal to (2, n_timestamps_1) "
      "(got {0})".format((1, 2)))]
)
def test_parameter_check_accumulated_cost_matrix(params, err_msg):
    """Test parameter validation."""
    with pytest.raises(ValueError, match=re.escape(err_msg)):
        accumulated_cost_matrix(**params)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'cost_mat': [[0, 2, 5], [2, 1, 4], [5, 4, 5]]},
      [[0, 2, 7], [2, 1, 5], [7, 5, 6]]),

     ({'cost_mat': [[0, 2, 5], [2, 1, 4], [5, 4, 5]],
       'region': [[0, 0, 1], [2, 3, 3]]},
      [[0, 2, np.inf], [2, 1, 5], [np.inf, 5, 6]])]
)
def test_actual_results_accumulated_cost_matrix(params, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = accumulated_cost_matrix(**params)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'acc_cost_mat, arr_desired',
    [(np.asarray([[0, 2, 7], [2, 1, 5], [7, 5, 6]]), [[0, 1, 2], [0, 1, 2]]),

     (np.asarray([[0, 0, 0], [2, 3, 1], [7, 5, 2]]),
      [[0, 0, 1, 2], [0, 1, 2, 2]]),

     (np.asarray([[0, 0, np.inf], [2, 3, 1], [np.inf, 5, 2]]),
      [[0, 0, 1, 2], [0, 1, 2, 2]])]
)
def test_actual_results_return_path(acc_cost_mat, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = _return_path(acc_cost_mat)
    np.testing.assert_array_equal(arr_actual, arr_desired)


@pytest.mark.parametrize(
    'params, res_desired',
    [({}, 2),

     ({'return_cost': True}, (2, [[2]])),

     ({'return_accumulated': True}, (2, [[3]])),

     ({'return_path': True}, (2, [[0], [0]])),

     ({'return_cost': True, 'return_accumulated': True},
      (2, [[2]], np.asarray([[3]]))),

     ({'return_cost': True, 'return_path': True}, (2, [[2]], [[0], [0]])),

     ({'return_accumulated': True, 'return_path': True},
      (2, np.asarray([[3]]), [[0], [0]])),

     ({'return_cost': True, 'return_accumulated': True,
       'return_path': True}, (2, [[2]], np.asarray([[3]]), [[0], [0]]))]
)
def test_actual_results_return_results(params, res_desired):
    """Test that the actual results are the expected ones."""
    res_actual = _return_results(dtw_dist=2, cost_mat=[[2]],
                                 acc_cost_mat=np.asarray([[3]]), **params)
    if isinstance(res_desired, tuple):
        for actual, desired in zip(res_actual, res_desired):
            np.testing.assert_allclose(actual, desired)
    else:
        np.testing.assert_allclose(res_actual, res_desired)


@pytest.mark.parametrize(
    'params, res_desired',
    [({},
      {'cost_mat': [[4, 0, 1], [1, 1, 0], [0, 4, 1]],
       'acc_cost_mat': [[4, 4, 5], [5, 5, 4], [5, 9, 5]],
       'path': [[0, 0, 1, 2], [0, 1, 2, 2]],
       'dtw': sqrt(5)}),

     ({'dist': 'absolute'},
      {'cost_mat': [[2, 0, 1], [1, 1, 0], [0, 2, 1]],
       'acc_cost_mat': [[2, 2, 3], [3, 3, 2], [3, 5, 3]],
       'path': [[0, 0, 1, 2], [0, 1, 2, 2]],
       'dtw': 3})]
)
def test_actual_results_dtw_classic(params, res_desired):
    """Test that the actual results are the expected ones."""
    (dtw_actual, cost_mat_actual,
     acc_cost_mat_actual, path_actual) = _dtw_classic(
        x, y, **params_return, **params
    )
    np.testing.assert_allclose(cost_mat_actual, res_desired['cost_mat'])
    np.testing.assert_allclose(dtw_actual, res_desired['dtw'])
    np.testing.assert_allclose(path_actual, res_desired['path'])
    np.testing.assert_allclose(acc_cost_mat_actual,
                               res_desired['acc_cost_mat'])


@pytest.mark.parametrize(
    'params, err_msg',
    [({'region': [[1, 1]]},
      "The shape of 'region' must be equal to (2, n_timestamps_1)")]
)
def test_parameter_check_dtw_region(params, err_msg):
    """Test parameter validation."""
    with pytest.raises(ValueError, match=re.escape(err_msg)):
        _dtw_region(x, y, **params)


@pytest.mark.parametrize(
    'params, res_desired',
    [({},
      {'cost_mat': [[4, 0, 1], [1, 1, 0], [0, 4, 1]],
       'acc_cost_mat': [[4, 4, 5], [5, 5, 4], [5, 9, 5]],
       'path': [[0, 0, 1, 2], [0, 1, 2, 2]],
       'dtw': sqrt(5)}),

     ({'dist': 'absolute'},
      {'cost_mat': [[2, 0, 1], [1, 1, 0], [0, 2, 1]],
       'acc_cost_mat': [[2, 2, 3], [3, 3, 2], [3, 5, 3]],
       'path': [[0, 0, 1, 2], [0, 1, 2, 2]],
       'dtw': 3}),

     ({'region': [[0, 0, 1], [2, 3, 3]]},
      {'cost_mat': [[4, 0, np.inf], [1, 1, 0], [np.inf, 4, 1]],
       'acc_cost_mat': [[4, 4, np.inf], [5, 5, 4], [np.inf, 9, 5]],
       'path': [[0, 0, 1, 2], [0, 1, 2, 2]],
       'dtw': sqrt(5)}),

     ({'dist': 'absolute', 'region': [[0, 0, 1], [2, 3, 3]]},
      {'cost_mat': [[2, 0, np.inf], [1, 1, 0], [np.inf, 2, 1]],
       'acc_cost_mat': [[2, 2, np.inf], [3, 3, 2], [np.inf, 5, 3]],
       'path': [[0, 0, 1, 2], [0, 1, 2, 2]],
       'dtw': 3})]
)
def test_actual_results_dtw_region(params, res_desired):
    """Test that the actual results are the expected ones."""
    (dtw_actual, cost_mat_actual,
     acc_cost_mat_actual, path_actual) = _dtw_region(
        x, y, **params_return, **params
    )
    np.testing.assert_allclose(cost_mat_actual, res_desired['cost_mat'])
    np.testing.assert_allclose(dtw_actual, res_desired['dtw'])
    np.testing.assert_allclose(path_actual, res_desired['path'])
    np.testing.assert_allclose(acc_cost_mat_actual,
                               res_desired['acc_cost_mat'])


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'n_timestamps_1': '3.'}, TypeError,
      "'n_timestamps_1' must be an integer."),

     ({'n_timestamps_1': 3, 'window_size': '0.5'}, TypeError,
      "'window_size' must be an integer or a float."),

     ({'n_timestamps_1': 1}, ValueError,
      "'n_timestamps_1' must be an integer greater than or equal to 2."),

     ({'n_timestamps_1': 10, 'window_size': 3.},
      ValueError, "The given 'window_size' is a float, "
                  "it must be between "
                  "0. and 1. To set the size of the sakoe-chiba "
                  "manually, 'window_size' must be an integer."),

     ({'n_timestamps_1': 10, 'window_size': 20},
      ValueError, "The given 'window_size' is an integer, it must "
                  "be greater "
                  "than or equal to 0 and lower than max('n_timestamps_1', "
                  "'n_timestamps_2')."),

     ({'n_timestamps_1': 10, 'window_size': "a"}, TypeError,
        "'window_size' must be an integer or a float.")]
)
def test_parameter_check_sakoe_chiba_band(params, error, err_msg):
    """Test parameter validation."""
    with pytest.raises(error, match=re.escape(err_msg)):
        sakoe_chiba_band(**params)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'n_timestamps_1': 4, 'n_timestamps_2': 4, 'window_size': 2},
     [[0, 0, 0, 1], [3, 4, 4, 4]]),
     ({'n_timestamps_1': 4, 'n_timestamps_2': 4, 'window_size': 0.5},
     [[0, 0, 0, 1], [3, 4, 4, 4]]),
     ({'n_timestamps_1': 4, 'n_timestamps_2': 4, 'window_size': 3},
     [[0, 0, 0, 0], [4, 4, 4, 4]]),
     ({'n_timestamps_1': 4, 'n_timestamps_2': 4, 'window_size': 1},
     [[0, 0, 1, 2], [2, 3, 4, 4]]),
     ({'n_timestamps_1': 4, 'n_timestamps_2': 4, 'window_size': 0},
     [[0, 1, 2, 3], [1, 2, 3, 4]]),
     ({'n_timestamps_1': 4, 'n_timestamps_2': 5, 'window_size': 2},
     [[0, 0, 1, 2], [3, 4, 5, 5]]),
     ({'n_timestamps_1': 5, 'n_timestamps_2': 4, 'window_size': 2},
     [[0, 0, 0, 1, 2], [2, 3, 4, 4, 4]]),
     ]
)
def test_actual_results_sakoe_chiba_band(params, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = sakoe_chiba_band(**params)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, res_desired',
    [({'window_size': 2},
      {'cost_mat': [[4, 0, 1], [1, 1, 0], [0, 4, 1]],
       'acc_cost_mat': [[4, 4, 5], [5, 5, 4], [5, 9, 5]],
       'path': [[0, 0, 1, 2], [0, 1, 2, 2]],
       'dtw': sqrt(5)}),

     ({'window_size': 2, 'dist': 'absolute'},
      {'cost_mat': [[2, 0, 1], [1, 1, 0], [0, 2, 1]],
       'acc_cost_mat': [[2, 2, 3], [3, 3, 2], [3, 5, 3]],
       'path': [[0, 0, 1, 2], [0, 1, 2, 2]],
       'dtw': 3}),

     ({'window_size': 1},
      {'cost_mat': [[4, 0, np.inf], [1, 1, 0], [np.inf, 4, 1]],
       'acc_cost_mat': [[4, 4, np.inf], [5, 5, 4], [np.inf, 9, 5]],
       'path': [[0, 0, 1, 2], [0, 1, 2, 2]],
       'dtw': sqrt(5)}),

     ({'window_size': 1, 'dist': 'absolute'},
      {'cost_mat': [[2, 0, np.inf], [1, 1, 0], [np.inf, 2, 1]],
       'acc_cost_mat': [[2, 2, np.inf], [3, 3, 2], [np.inf, 5, 3]],
       'path': [[0, 0, 1, 2], [0, 1, 2, 2]],
       'dtw': 3}),

     ({'window_size': 0},
      {'cost_mat':
          [[4, np.inf, np.inf], [np.inf, 1, np.inf], [np.inf, np.inf, 1]],
       'acc_cost_mat':
           [[4, np.inf, np.inf], [np.inf, 5, np.inf], [np.inf, np.inf, 6]],
       'path': [[0, 1, 2], [0, 1, 2]],
       'dtw': sqrt(6)})]
)
def test_actual_results_dtw_sakoechiba(params, res_desired):
    """Test that the actual results are the expected ones."""
    (dtw_actual, cost_mat_actual,
     acc_cost_mat_actual, path_actual) = _dtw_sakoechiba(
        x, y, **params_return, **params
    )
    np.testing.assert_allclose(cost_mat_actual, res_desired['cost_mat'])
    np.testing.assert_allclose(dtw_actual, res_desired['dtw'])
    np.testing.assert_allclose(path_actual, res_desired['path'])
    np.testing.assert_allclose(acc_cost_mat_actual,
                               res_desired['acc_cost_mat'])


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'n_timestamps_1': '3'}, TypeError,
      "'n_timestamps_1' must be an integer."),

     ({'n_timestamps_1': 10, 'n_timestamps_2': 10, 'max_slope': '3'},
     TypeError,
      "'max_slope' must be an integer or a float."),

     ({'n_timestamps_1': 1}, ValueError,
      "'n_timestamps_1' must be an integer greater than or equal to 2."),

     ({'n_timestamps_1': 10, 'max_slope': 0.5}, ValueError,
      "'max_slope' must be a number greater than or equal to 1.")]
)
def test_parameter_check_itakura_parallelogram(params, error, err_msg):
    """Test parameter validation."""
    with pytest.raises(error, match=re.escape(err_msg)):
        itakura_parallelogram(**params)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'n_timestamps_1': 4, 'n_timestamps_2': 4, 'max_slope': 2},
     [[0, 1, 1, 3], [1, 3, 3, 4]]),
     ({'n_timestamps_1': 4, 'n_timestamps_2': 4, 'max_slope': 8},
     [[0, 1, 1, 3], [1, 3, 3, 4]]),
     ({'n_timestamps_1': 4, 'n_timestamps_2': 4, 'max_slope': 1},
     [[0, 1, 2, 3], [1, 2, 3, 4]]),
     ({'n_timestamps_1': 6, 'n_timestamps_2': 3, 'max_slope': 1},
     [[0, 1, 1, 1, 1, 2], [1, 2, 2, 2, 2, 3]]),
     ({'n_timestamps_1': 4, 'n_timestamps_2': 6, 'max_slope': 1},
     [[0, 1, 3, 5], [1, 3, 5, 6]])]
)
def test_actual_results_itakura_parallelogram(params, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = itakura_parallelogram(**params)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, res_desired',
    [({},
      {'cost_mat':
          [[4, np.inf, np.inf], [np.inf, 1, np.inf], [np.inf, np.inf, 1]],
       'acc_cost_mat':
           [[4, np.inf, np.inf], [np.inf, 5, np.inf], [np.inf, np.inf, 6]],
       'path': [[0, 1, 2], [0, 1, 2]],
       'dtw': sqrt(6)}),

     ({'max_slope': 1, 'dist': 'absolute'},
      {'cost_mat':
          [[2, np.inf, np.inf], [np.inf, 1, np.inf], [np.inf, np.inf, 1]],
       'acc_cost_mat':
          [[2, np.inf, np.inf], [np.inf, 3, np.inf], [np.inf, np.inf, 4]],
       'path': [[0, 1, 2], [0, 1, 2]],
       'dtw': 4}),

     ({'max_slope': 8},
      {'cost_mat':
          [[4, np.inf, np.inf], [np.inf, 1, np.inf], [np.inf, np.inf, 1]],
       'acc_cost_mat':
          [[4, np.inf, np.inf], [np.inf, 5, np.inf], [np.inf, np.inf, 6]],
       'path': [[0, 1, 2], [0, 1, 2]],
       'dtw': sqrt(6)}),

     ({'max_slope': 8, 'dist': 'absolute'},
      {'cost_mat':
          [[2, np.inf, np.inf], [np.inf, 1, np.inf], [np.inf, np.inf, 1]],
       'acc_cost_mat':
          [[2, np.inf, np.inf], [np.inf, 3, np.inf], [np.inf, np.inf, 4]],
       'path': [[0, 1, 2], [0, 1, 2]],
       'dtw': 4})]
)
def test_actual_results_dtw_itakura(params, res_desired):
    """Test that the actual results are the expected ones."""
    (dtw_actual, cost_mat_actual,
     acc_cost_mat_actual, path_actual) = _dtw_itakura(
        x, y, **params_return, **params
    )
    np.testing.assert_allclose(cost_mat_actual, res_desired['cost_mat'])
    np.testing.assert_allclose(dtw_actual, res_desired['dtw'])
    np.testing.assert_allclose(path_actual, res_desired['path'])
    np.testing.assert_allclose(acc_cost_mat_actual,
                               res_desired['acc_cost_mat'])


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'n_timestamps_1': 6, 'n_timestamps_2': 6, 'resolution_level': 2,
     'n_timestamps_reduced_1': 3, 'n_timestamps_reduced_2': 3,
       'path': np.asarray([[0, 0, 1, 2], [0, 1, 2, 2]]), 'radius': 0},
      [[0, 0, 4, 4, 4, 4], [4, 4, 6, 6, 6, 6]]),

     ({'n_timestamps_1': 6, 'n_timestamps_2': 6, 'resolution_level': 2,
      'n_timestamps_reduced_1': 3, 'n_timestamps_reduced_2': 3,
       'path': np.asarray([[0, 0, 1, 2], [0, 1, 2, 2]]), 'radius': 1},
      [[0, 0, 0, 0, 2, 2], [6, 6, 6, 6, 6, 6]]),
     ({'n_timestamps_1': 6, 'n_timestamps_2': 4, 'resolution_level': 2,
      'n_timestamps_reduced_1': 2, 'n_timestamps_reduced_2': 2,
       'path': np.asarray([[0, 0, 1, 2], [0, 1, 2, 2]]), 'radius': 1},
      [[0, 0, 0, 0], [4, 4, 4, 4]])
     ]
)
def test_actual_results_blurred_path_region(params, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = _blurred_path_region(**params)
    np.testing.assert_array_equal(arr_actual, arr_desired)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'resolution': '3'}, TypeError, "'resolution' must be an integer."),
     ({'radius': '3'}, TypeError, "'radius' must be an integer."),
     ({'resolution': 0}, ValueError,
      "'resolution' must be a positive integer."),
     ({'radius': -3}, ValueError, "'radius' must be a non-negative integer.")]
)
def test_parameter_check_dtw_multiscale(params, error, err_msg):
    """Test parameter validation."""
    with pytest.raises(error, match=re.escape(err_msg)):
        _dtw_multiscale(x, y, **params)


@pytest.mark.parametrize(
    'params, res_desired',
    [({},
      {'cost_mat':
          [[1, 4, np.inf, np.inf], [0, 1, np.inf, np.inf],
           [np.inf, np.inf, 1, 4], [np.inf, np.inf, 0, 1]],
       'acc_cost_mat':
           [[1, 5, np.inf, np.inf], [1, 2, np.inf, np.inf],
            [np.inf, np.inf, 3, 7], [np.inf, np.inf, 3, 4]],
       'path': [[0, 1, 2, 3], [0, 1, 2, 3]],
       'dtw': 2}),

     ({'radius': 1},
      {'cost_mat':
          [[1, 4, 9, 16], [0, 1, 4, 9],
           [1, 0, 1, 4], [4, 1, 0, 1]],
       'acc_cost_mat':
           [[1, 5, 14, 30], [1, 2, 6, 15],
            [2, 1, 2, 6], [6, 2, 1, 2]],
       'path': [[0, 1, 2, 3, 3], [0, 0, 1, 2, 3]],
       'dtw': sqrt(2)}),

     ({'radius': 2},
      {'cost_mat':
          [[1, 4, 9, 16], [0, 1, 4, 9],
           [1, 0, 1, 4], [4, 1, 0, 1]],
       'acc_cost_mat':
           [[1, 5, 14, 30], [1, 2, 6, 15],
            [2, 1, 2, 6], [6, 2, 1, 2]],
       'path': [[0, 1, 2, 3, 3], [0, 0, 1, 2, 3]],
       'dtw': sqrt(2)})]
)
def test_actual_results_dtw_multiscale(params, res_desired):
    """Test that the actual results are the expected ones."""
    x = np.arange(4)
    y = np.arange(1, 5)

    (dtw_actual, cost_mat_actual,
     acc_cost_mat_actual, path_actual) = _dtw_multiscale(
        x, y, **params_return, **params
    )
    np.testing.assert_allclose(cost_mat_actual, res_desired['cost_mat'])
    np.testing.assert_allclose(dtw_actual, res_desired['dtw'])
    np.testing.assert_allclose(path_actual, res_desired['path'])
    np.testing.assert_allclose(acc_cost_mat_actual,
                               res_desired['acc_cost_mat'])


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'radius': '3'}, TypeError, "'radius' must be an integer."),
     ({'radius': -3}, ValueError, "'radius' must be a non-negative integer.")]
)
def test_parameter_check_dtw_fast(params, error, err_msg):
    """Test parameter validation."""
    with pytest.raises(error, match=re.escape(err_msg)):
        _dtw_fast(x, y, **params)


@pytest.mark.parametrize(
    'params, res_desired',
    [({},
      {'cost_mat':
          [[1, 4, np.inf, np.inf], [0, 1, np.inf, np.inf],
           [np.inf, np.inf, 1, 4], [np.inf, np.inf, 0, 1]],
       'acc_cost_mat':
           [[1, 5, np.inf, np.inf], [1, 2, np.inf, np.inf],
            [np.inf, np.inf, 3, 7], [np.inf, np.inf, 3, 4]],
       'path': [[0, 1, 2, 3], [0, 1, 2, 3]],
       'dtw': 2}),

     ({'radius': 1},
      {'cost_mat':
          [[1, 4, 9, 16], [0, 1, 4, 9],
           [1, 0, 1, 4], [4, 1, 0, 1]],
       'acc_cost_mat':
           [[1, 5, 14, 30], [1, 2, 6, 15],
            [2, 1, 2, 6], [6, 2, 1, 2]],
       'path': [[0, 1, 2, 3, 3], [0, 0, 1, 2, 3]],
       'dtw': sqrt(2)}),

     ({'radius': 2},
      {'cost_mat':
          [[1, 4, 9, 16], [0, 1, 4, 9],
           [1, 0, 1, 4], [4, 1, 0, 1]],
       'acc_cost_mat':
           [[1, 5, 14, 30], [1, 2, 6, 15],
            [2, 1, 2, 6], [6, 2, 1, 2]],
       'path': [[0, 1, 2, 3, 3], [0, 0, 1, 2, 3]],
       'dtw': sqrt(2)})]
)
def test_actual_results_dtw_fast(params, res_desired):
    """Test that the actual results are the expected ones."""
    x = np.arange(4)
    y = np.arange(1, 5)

    (dtw_actual, cost_mat_actual,
     acc_cost_mat_actual, path_actual) = _dtw_fast(
        x, y, **params_return, **params
    )
    np.testing.assert_allclose(cost_mat_actual, res_desired['cost_mat'])
    np.testing.assert_allclose(dtw_actual, res_desired['dtw'])
    np.testing.assert_allclose(path_actual, res_desired['path'])
    np.testing.assert_allclose(acc_cost_mat_actual,
                               res_desired['acc_cost_mat'])


@pytest.mark.parametrize(
    'params, err_msg',
    [({'method': 'loop'},
      "'method' must be either 'classic', 'sakoechiba', 'itakura', "
      "'multiscale' or 'fast'.")]
)
def test_parameter_check_dtw(params, err_msg):
    """Test parameter validation."""
    with pytest.raises(ValueError, match=re.escape(err_msg)):
        dtw(x, y, **params)


@pytest.mark.parametrize(
    'params, res_desired',
    [({}, _dtw_classic(x, y, **params_return)),

     ({'method': 'sakoechiba'}, _dtw_sakoechiba(x, y, **params_return)),

     ({'method': 'sakoechiba', 'options': {'window_size': 0.5}},
      _dtw_sakoechiba(x, y, **params_return)),

     ({'method': 'itakura'}, _dtw_itakura(x, y, **params_return)),

     ({'method': 'itakura', 'options': {'max_slope': 8}},
      _dtw_itakura(x, y, max_slope=8, **params_return)),

     ({'method': 'multiscale'}, _dtw_multiscale(x, y, **params_return)),

     ({'method': 'multiscale', 'options': {'resolution': 1}},
      _dtw_multiscale(x, y, resolution=1, **params_return)),

     ({'method': 'multiscale', 'options': {'radius': 2}},
      _dtw_multiscale(x, y, radius=2, **params_return)),

     ({'method': 'fast'}, _dtw_fast(x, y, **params_return)),

     ({'method': 'fast', 'options': {'radius': 1}},
      _dtw_fast(x, y, radius=1, **params_return))]
)
def test_actual_results_dtw(params, res_desired):
    """Test that the actual results are the expected ones."""
    (dtw_actual, cost_mat_actual, acc_cost_mat_actual, path_actual) = dtw(
        x, y, **params_return, **params
    )
    np.testing.assert_allclose(dtw_actual, res_desired[0])
    np.testing.assert_allclose(cost_mat_actual, res_desired[1])
    np.testing.assert_allclose(acc_cost_mat_actual, res_desired[2])
    np.testing.assert_allclose(path_actual, res_desired[3])


@pytest.mark.parametrize(
    'params, res_desired',
    [({}, _dtw_classic(x_long, y, **params_return)),
     ({}, _dtw_classic(x_long, y, **params_return)),
     ({'method': 'sakoechiba'}, _dtw_sakoechiba(x_long, y, **params_return)),

     ({'method': 'sakoechiba', 'options': {'window_size': 0.5}},
      _dtw_sakoechiba(x_long, y, window_size=0.5, **params_return)),

     ({'method': 'itakura'}, _dtw_itakura(x_long, y, **params_return)),

     ({'method': 'itakura', 'options': {'max_slope': 8}},
      _dtw_itakura(x_long, y, max_slope=8, **params_return)),

     ({'method': 'multiscale'}, _dtw_multiscale(x_long, y, **params_return)),

     ({'method': 'multiscale', 'options': {'resolution': 1}},
      _dtw_multiscale(x_long, y, resolution=1, **params_return)),

     ({'method': 'multiscale', 'options': {'radius': 2}},
      _dtw_multiscale(x_long, y, radius=2, **params_return)),

     ({'method': 'fast'}, _dtw_fast(x_long, y, **params_return)),

     ({'method': 'fast', 'options': {'radius': 1}},
      _dtw_fast(x_long, y, radius=1, **params_return))]
)
def test_actual_results_dtw_diff_lengths(params, res_desired):
    """Test that the actual results are the expected ones."""
    (dtw_actual, cost_mat_actual, acc_cost_mat_actual, path_actual) = dtw(
        x_long, y, **params_return, **params
    )
    np.testing.assert_allclose(dtw_actual, res_desired[0])
    np.testing.assert_allclose(cost_mat_actual, res_desired[1])
    np.testing.assert_allclose(acc_cost_mat_actual, res_desired[2])
    np.testing.assert_allclose(path_actual, res_desired[3])


@pytest.mark.parametrize(
    "region, n_timestamps_1, n_timestamps_2",
    [(np.array([[-3, 0, 1, 2, 10, 30], [0, 2, 4, 5, 10, 50]]), 4, 10),
     (np.array([[-3, 0, 1, 2, 10, 30], [0, 2, 4, 5, 10, 50]]), 6, 3)])
def test_check_region(region, n_timestamps_1, n_timestamps_2):
    region = _check_region(region, n_timestamps_1, n_timestamps_2)
    assert region.min() >= 0
    assert region.max() <= n_timestamps_2
    assert region.shape[1] == n_timestamps_1


@pytest.mark.parametrize('method', ['classic', 'sakoechiba', 'itakura'])
@pytest.mark.parametrize('n_timestamps_1, n_timestamps_2', [[4, 10], [4, 10]])
def test_precomputed_cost(method, n_timestamps_1, n_timestamps_2):
    rng = np.random.RandomState(42)
    x = rng.randn(n_timestamps_1)
    y = rng.randn(n_timestamps_2)
    squared_cost = (x[:, None] - y[None, :]) ** 2
    dtw_desired = dtw(x, y, method=method, dist="square")
    dtw_desired **= 2
    dtw_precomputed = dtw(method=method, dist="precomputed",
                          precomputed_cost=squared_cost)
    np.testing.assert_allclose(dtw_desired, dtw_precomputed)


@pytest.mark.parametrize("method", ["fast", "multiscale"])
def test_precomputed_cost_error(method):
    error_match = "cannot be used with a precomputed cost"
    with pytest.raises(ValueError, match=error_match):
        dtw(method=method, dist="precomputed")


@pytest.mark.parametrize(
    "method, dtw_func",
    [("classic", _dtw_classic),
     ("multiscale", _dtw_multiscale),
     ("itakura", _dtw_itakura),
     ("sakoechiba", _dtw_sakoechiba),
     ("fast", _dtw_fast)])
def test_dtw_methods(method, dtw_func):
    value_specific = dtw_func(x, y)
    value_generic = dtw(x, y, method=method)
    assert value_specific == value_generic


@pytest.mark.parametrize(
    'params', [{'method': 'whoops'}, {'method': 'sakoe'}]
)
def test_parameter_check_show_options(params):
    """Test parameter validation."""
    err_msg = ("'method' must be either None, 'classic', 'sakoechiba', "
               "'itakura', 'region', 'multiscale' or 'fast'.")
    with pytest.raises(ValueError, match=re.escape(err_msg)):
        show_options(**params)


@pytest.mark.parametrize(
    'params',
    [{},
     {'disp': False},
     {'method': 'classic'},
     {'method': 'classic', 'disp': False},
     {'method': 'sakoechiba'},
     {'method': 'sakoechiba', 'disp': False},
     {'method': 'itakura'},
     {'method': 'itakura', 'disp': False},
     {'method': 'region'},
     {'method': 'region', 'disp': False},
     {'method': 'multiscale'},
     {'method': 'multiscale', 'disp': False},
     {'method': 'fast'},
     {'method': 'fast', 'disp': False}]
)
def test_actual_results_show_options(params):
    """Test that the actual results are the expected ones."""
    res_actual = show_options(**params)
    if ('disp' in params.keys()) and (not params['disp']):
        assert isinstance(res_actual, str)
    else:
        assert res_actual is None



"""Testing for transformers."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from scipy.stats import boxcox, norm, yeojohnson
from pyts.preprocessing import StandardScaler
from pyts.preprocessing import PowerTransformer, QuantileTransformer


X = np.arange(1, 34).reshape(3, 11)


def test_actual_results_power_transformer_box_cox():
    """Test that the actual results are the expected ones."""
    for standardize in [True, False]:
        pt = PowerTransformer(method='box-cox', standardize=standardize)
        arr_actual = pt.fit_transform(X)
        arr_desired = [boxcox(X[i])[0] for i in range(3)]
        if standardize:
            arr_desired = StandardScaler().transform(arr_desired)
        np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


def test_actual_results_power_transformer_yeo_johnson():
    """Test that the actual results are the expected ones."""
    for standardize in [True, False]:
        pt = PowerTransformer(method='yeo-johnson', standardize=standardize)
        arr_actual = pt.fit_transform(X)
        arr_desired = [yeojohnson(X[i].astype('float64'))[0] for i in range(3)]
        if standardize:
            arr_desired = StandardScaler().transform(arr_desired)
        np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


def test_actual_results_quantile_transformer_uniform():
    """Test that the actual results are the expected ones."""
    transformer = QuantileTransformer(n_quantiles=11)
    arr_actual = transformer.fit_transform(X)
    arr_desired = [np.linspace(0, 1, 11) for _ in range(3)]
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


def test_actual_results_quantile_transformer_normal():
    """Test that the actual results are the expected ones."""
    X_ppf = norm.ppf(np.linspace(0, 1, 1000)[1:-1])
    weights = np.round(norm.pdf(X_ppf) * 1000).astype('int64')
    X = []
    for value, weight in zip(X_ppf, weights):
        X += [value] * weight
    X = np.asarray(X).reshape(1, -1)
    transformer = QuantileTransformer(n_quantiles=11,
                                      output_distribution='normal')
    arr_actual = transformer.fit_transform(X)
    arr_desired = X
    atol = 0.01 * X.shape[1]
    np.testing.assert_allclose(arr_actual, arr_desired, atol=atol, rtol=0.)



"""Testing for MultivariateClassifier."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.classification import SAXVSM, BOSSVS
from pyts.multivariate.classification import MultivariateClassifier
from pyts.transformation import BOSS


n_samples, n_features, n_timestamps, n_classes = 40, 3, 30, 2
rng = np.random.RandomState(42)
X = rng.randn(n_samples, n_features, n_timestamps)
y = rng.randint(n_classes, size=n_samples)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'estimator': [SAXVSM(), SAXVSM(), BOSS()]},
      ValueError, "Estimator 2 must be a classifier."),

     ({'estimator': [SAXVSM()]}, ValueError,
      "If 'estimator' is a list, its length must be equal to "
      "the number of features (1 != 3)"),

     ({'estimator': None}, TypeError,
      "'estimator' must be a classifier that inherits from "
      "sklearn.base.BaseEstimator or a list thereof.")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    clf = MultivariateClassifier(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        clf.fit(X, y)


@pytest.mark.parametrize(
    'params, X, error, err_msg',
    [({'estimator': BOSS()}, X[0, 0], ValueError,
      "X must be 3-dimensional (got 1)."),
     ({'estimator': BOSS()}, X[0], ValueError,
      "X must be 3-dimensional (got 2).")]
)
def test_input_check(params, X, error, err_msg):
    """Test input data validation."""
    clf = MultivariateClassifier(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        clf.fit(X, y)


@pytest.mark.parametrize(
    'params',
    [{'estimator': SAXVSM()},
     {'estimator': [SAXVSM() for _ in range(n_features)]},
     {'estimator': [SAXVSM(), SAXVSM(), BOSSVS()]}]
)
def test_actual_results_without_weights(params):
    """Test that the actual results are the expected ones."""
    arr_actual = MultivariateClassifier(**params).fit(X, y).predict(X)
    predictions = []
    if isinstance(params['estimator'], list):
        for i in range(n_features):
            predictions.append(
                params['estimator'][i].fit(X[:, i], y).predict(X[:, i]))
    else:
        for i in range(n_features):
            predictions.append(
                params['estimator'].fit(X[:, i], y).predict(X[:, i]))
    predictions = np.asarray(predictions)
    arr_desired = []
    for i in range(n_samples):
        arr_desired.append(np.argmax(np.bincount(predictions[:, i])))
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params',
    [{'estimator': SAXVSM(), 'weights': [0.1, 0.7, 0.2]},

     {'estimator': [SAXVSM() for _ in range(n_features)],
      'weights': [0.1, 0.7, 0.2]},

     {'estimator': [SAXVSM(), SAXVSM(), BOSSVS()],
      'weights': [0.1, 0.7, 0.2]}]
)
def test_actual_results_with_weights(params):
    """Test that the actual results are the expected ones."""
    arr_actual = MultivariateClassifier(**params).fit(X, y).predict(X)
    predictions = []
    if isinstance(params['estimator'], list):
        for i in range(n_features):
            predictions.append(
                params['estimator'][i].fit(X[:, i], y).predict(X[:, i]))
    else:
        for i in range(n_features):
            predictions.append(
                params['estimator'].fit(X[:, i], y).predict(X[:, i]))
    predictions = np.asarray(predictions)
    arr_desired = []
    for i in range(n_samples):
        arr_desired.append(np.argmax(
            np.bincount(predictions[:, i], params['weights'])))
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)
